from fusion.libs.state import view_state_type
from pamet.model.note import Note
from pamet.views.note.base.state import NoteViewState


@view_state_type
class NoteEditViewState(NoteViewState):
    new_note_dict: dict = None

    @property
    def create_mode(self):
        return bool(self.new_note_dict)

    def update_from_note(self, note: Note):
        note_dict = note.asdict()
        note_dict.pop('id')
        self.replace(**note_dict)



from __future__ import annotations
from importlib import resources
from pathlib import Path

from typing import Union
from fusion.util.point2d import Point2D
from fusion.logging import get_logger
import pamet
from pamet.constants import ALIGNMENT_GRID_UNIT
from pamet.model.page import Page

log = get_logger(__name__)


def resource_path(subpath: Union[str, Path]):
    # resource_dir_path = Path(__file__).parent.parent / 'resources'
    # resource_path = resource_dir_path / Path(subpath)
    subpath = Path(subpath)
    # resource_path = resources.path('pamet.resources', subpath)
    resource_path = resource_dir(subpath.parent) / subpath.name

    # if subpath.startswith('/'):
    #     subpath = subpath[1:]

    if not resource_path.exists():
        raise Exception('Resource not found')
    return resource_path


def resource_dir(dir_subpath: str | Path):
    resources_folder = resources.files('pamet.resources')
    # if not dir_subpath:
    #     return resources_folder
    dir_subpath = Path(dir_subpath)
    return resources_folder / dir_subpath


def snap_to_grid(x: Union[float, Point2D]) -> Union[float, Point2D]:
    return round(x / ALIGNMENT_GRID_UNIT) * ALIGNMENT_GRID_UNIT


def generate_page_name() -> str:
    base_name = 'New page'  # Could be replaced with config, translations, etc.

    page_name = base_name
    default_names_found = 0
    while True:
        if not pamet.find_one(name=page_name, type=Page):
            return page_name
        default_names_found += 1
        page_name = f'{base_name} {default_names_found}'


# Validity checks
def arrow_validity_check():
    """Removes arrows with missing anchor notes. Must be performed after
    the repo has been initialized."""
    from pamet.model.arrow import Arrow

    # Validity check on the arrow note anchors
    # If the note, that the anchor points to, is missing - skip arrow
    invalid_arrows = []
    for arrow in pamet.find(type=Arrow):
        tail_note = pamet.note(arrow.page_id, arrow.get_tail_note_own_id())
        head_note = pamet.note(arrow.page_id, arrow.get_head_note_own_id())
        if (arrow.has_tail_anchor() and not tail_note) or\
                (arrow.has_head_anchor() and not head_note):

            invalid_arrows.append(arrow)

    for arrow in invalid_arrows:
        log.error('Removing arrow because of an invalid note anchor: '
                  f'{arrow}')
        pamet.remove_arrow(arrow)



from PySide6.QtWidgets import QLabel
from PySide6.QtGui import QPainter, QColor
from PySide6.QtCore import Qt, QRect

from fusion.platform.qt_widgets import bind_and_apply_state
from fusion.libs.state import view_state_type

import pamet
from pamet import register_note_view_type
from pamet.desktop_app import default_note_font
from pamet.model.text_note import TextNote
from pamet.views.note.anchor.view_mixin import LinkNoteViewMixin
from pamet.views.note.base.view import NoteView
from pamet.views.note.base.state import NoteViewState
from pamet.desktop_app.util import elide_text, draw_text_lines

from fusion import get_logger
from pamet.views.note.qt_helpers import draw_link_decorations

log = get_logger(__name__)


@view_state_type
class TextNoteViewState(NoteViewState, TextNote):
    pass


@register_note_view_type(state_type=TextNoteViewState,
                         note_type=TextNote,
                         edit=False)
class TextNoteWidget(QLabel, NoteView, LinkNoteViewMixin):

    def __init__(self, parent, initial_state):
        QLabel.__init__(self, parent=parent)
        NoteView.__init__(self, parent=parent, initial_state=initial_state)
        LinkNoteViewMixin.__init__(self)

        self.setFont(default_note_font())

        self._elided_text_layout = None
        self._alignment = Qt.AlignHCenter
        self.setMargin(0)

        bind_and_apply_state(self, initial_state, self.on_state_change)
        # If there's a link - we watch for a page rename (and must unsubscribe
        # at destroyed)
        self.destroyed.connect(lambda: self.disconnect_from_page_changes())

    def on_state_change(self, change):
        state = change.last_state()
        if change.updated.color or change.updated.background_color:

            fg_col = QColor(*state.get_color().to_uint8_rgba_list())
            bg_col = QColor(*state.get_background_color().to_uint8_rgba_list())

            palette = self.palette()
            palette.setColor(self.backgroundRole(), bg_col)
            palette.setColor(self.foregroundRole(), fg_col)
            self.setPalette(palette)

        if change.updated.geometry:
            self.setGeometry(QRect(*state.geometry))

        if change.updated.text or \
                change.updated.geometry \
                or change.updated.url:
            # font = self.font()
            # font.setPixelSize(20)
            # font.setPointSizeF(note_props['font_size'] * font.pointSizeF())
            # font.setPointSizeF(14)
            # self.setFont(font)

            # font_metrics = QFontMetrics(self.font())
            # print('Font ascent', font_metrics.ascent())
            # print('Font descent', font_metrics.descent())
            # print('Font height', font_metrics.height())
            # print('Font leading', font_metrics.leading())
            # print('Font lineSpacing', font_metrics.lineSpacing())
            # print('Font pointSizeF', self.font().pointSizeF())

            # self.setText('<p style="line-height:%s%%;margin-top:-5px;margin-right
            # :5px;margin-bottom:5px">%s</p>' %
            #      (100*20/float(font_metrics.lineSpacing()), _elided_text_layout))
            # self.setText(elided_text)

            if '\n' in state.text:
                self._alignment = Qt.AlignLeft
            else:
                self._alignment = Qt.AlignHCenter

            self.recalculate_elided_text()

        if change.updated.url:
            url_page = pamet.page(state.url.get_page_id())
            if url_page and url_page.name:
                self.connect_to_page_changes(url_page.id)
            else:
                self.disconnect_from_page_changes()

    def paintEvent(self, event):
        painter = QPainter()
        painter.begin(self)

        draw_text_lines(painter, self._elided_text_layout.data,
                        self._alignment,
                        self.state().text_rect())
        draw_link_decorations(self, painter)
        painter.end()



from __future__ import annotations
import hashlib
import io
from typing import List, Tuple
from pathlib import Path
import shutil

from PIL import Image

from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QFont, QGuiApplication, QFontMetrics, QImage
from PySide6.QtGui import QImageReader, QPainter
from PySide6.QtCore import QBuffer, QByteArray, QIODevice, Qt
from PySide6.QtCore import QRectF, QRect

from fusion.util.rectangle import Rectangle
from fusion.logging import get_logger

import pamet
from pamet.constants import NO_SCALE_LINE_SPACING
from pamet.util import resource_dir

log = get_logger(__name__)


def control_is_pressed():
    return QGuiApplication.queryKeyboardModifiers() & Qt.ControlModifier


def shift_is_pressed():
    return QGuiApplication.queryKeyboardModifiers() & Qt.ShiftModifier


def current_window() -> WindowWidget:
    return QApplication.activeWindow()


def current_tab() -> TabWidget:
    return current_window().current_tab()


class TextLayout:

    def __init__(self):
        self.data = []
        self.is_elided = False

    def text(self):
        return '\n'.join([text for text, line in self.data])


def elide_text(text, text_rect: Rectangle, font: QFont) -> TextLayout:
    font_metrics = QFontMetrics(font)

    # Get the needed parameters
    line_spacing = NO_SCALE_LINE_SPACING

    # Get the y coordinates of the lines
    line_vpositions = []
    line_y = text_rect.top()
    while line_y <= text_rect.bottom() - line_spacing:
        line_vpositions.append(line_y)
        line_y += line_spacing

    # Divide the text into words and "mark" the ones ending with am
    # EoL char (by keeping their indexes in eol_word_indices)
    words = []
    eol_word_indices = []

    for line in text.split('\n'):
        words_on_line = line.split()
        words.extend(words_on_line)
        eol_word_indices.append(len(words) - 1)

    # Start filling the available lines one by one
    text_layout = TextLayout()

    # In case for some reason the text rect is too small to fit any text
    if not line_vpositions and text:
        text_layout.is_elided = True
        return text_layout

    ellide_line_end = False
    word_reached_idx = 0
    ellipsis_width = font_metrics.boundingRect('...').width()

    for line_idx, line_y in enumerate(line_vpositions):
        words_left = words[word_reached_idx:]

        # Find the coordinates and dimentions of the line
        line_rect = QRectF(*text_rect.as_tuple())
        line_rect.moveTop(line_y)
        line_rect.setHeight(line_spacing)

        # Fill the line word by word
        words_on_line = []
        width_left = text_rect.width()
        used_words = 0

        for word_idx_on_line, word in enumerate(words_left):
            # Add a leading space except before the first word
            if word_idx_on_line != 0:
                word = ' ' + word

            at_the_last_line = line_idx == (len(line_vpositions) - 1)
            at_last_word = word_idx_on_line == (len(words_left) - 1)

            # Get the dimentions of the word if drawn
            word_bbox = font_metrics.boundingRect(word)

            # There's enough space on the line for the next word
            if width_left >= word_bbox.width():
                width_left -= word_bbox.width()
                words_on_line.append(word)
                used_words += 1

            else:  # There's not enough space for the next word

                # If there's no room to add an elided word - ellide the
                # previous
                if at_the_last_line and width_left < ellipsis_width:
                    # words_on_line[-1] = words_on_line[-1]
                    ellide_line_end = True

                # Elide if we're past the end of the last line
                # or if it's the first word on the line and it's just too long
                if at_the_last_line or word_idx_on_line == 0:
                    word = font_metrics.elidedText(word, Qt.ElideRight,
                                                   width_left)
                    text_layout.is_elided = True
                    words_on_line.append(word)
                    used_words += 1

                # Done with this line - break and start the next if any
                break

            # Check if we're on EoL (because of a line break in the text)
            if (word_reached_idx + word_idx_on_line) in eol_word_indices:
                if at_the_last_line and not at_last_word:
                    ellide_line_end = True
                break

        if not words_on_line:
            break

        word_reached_idx += used_words
        line_text = ''.join(words_on_line)

        if ellide_line_end:
            text_layout.is_elided = True
            if len(line_text) >= 3:
                line_text = line_text[:-3] + '...'
            else:
                line_text = '...'

        text_layout.data.append((line_text, line_rect))
    return text_layout


def draw_text_lines(painter: QPainter, text_layout: List[Tuple[str, QRectF]],
                    alignment: Qt.AlignmentFlag, text_rect: QRect):
    if not text_layout:
        return
    # Qt configures the fonts with a lot of ascent
    # which makes them appear not properly centered
    hacky_padding = 8

    text_rect_height = text_rect.height()
    first_rect = text_layout[0][1]
    last_rect = text_layout[-1][1]
    text_height = last_rect.bottom() - first_rect.top()
    vertical_offset = (text_rect_height - text_height) / 2
    vertical_offset -= hacky_padding / 2

    for line_text, elide_line_rect in text_layout:
        line_rect = QRectF(elide_line_rect)
        line_rect.moveTop(vertical_offset + line_rect.top())
        line_rect.setHeight(line_rect.height() + hacky_padding)
        after_rect = painter.drawText(line_rect, alignment | Qt.TextDontClip,
                                      line_text)

        # print(after_rect)
        # painter.drawRect(line_rect)
        # painter.drawRect(after_rect)


def copy_script_templates(overwrite: bool = False):
    user_settings = pamet.desktop_app.get_user_settings()
    templates_folder: Path = Path(user_settings.script_templates_folder)
    templates_folder.mkdir(parents=True, exist_ok=True)
    source_folder = resource_dir('script_templates')

    for source_file in source_folder.iterdir():
        if not source_file.is_file():
            continue

        target_path = templates_folder / source_file.name
        if target_path.exists() and not overwrite:
            continue

        shutil.copy(source_file, target_path)


def convert_image_with_pil(image_data: bytes):
    bytes_io = io.BytesIO(image_data)
    img_byte_arr = io.BytesIO()

    try:  # Probe if Pillow can handle the image
        pil_image = Image.open(bytes_io)  #@IgnoreException
        pil_image.save(img_byte_arr, 'png')
    except Exception as e:
        raise Exception(f'Could not read image with PIL: "{e}"')

    return img_byte_arr.read()


def get_image_and_md5_from_bytearray(
        bytearray: QByteArray) -> Tuple[QImage, str]:
    if not isinstance(bytearray, QByteArray) or bytearray.isEmpty():
        raise Exception('Bytearray is wrong type or empty')

    # create buffer
    image_data = bytearray.data()
    buffer = QBuffer(bytearray)
    image_reader = QImageReader(buffer)

    # If Qt doesn't support the format (e.g. WebP)
    # try to convert it with Pillow to PNG
    if image_reader.format() not in \
            image_reader.supportedImageFormats():

        image_data = convert_image_with_pil(image_data)
        if not image_data:
            return QImage(), None
        buffer = QBuffer(QByteArray(image_data))
        image_reader = QImageReader(buffer)

    image = image_reader.read()
    if image.isNull():
        raise Exception(
            f'Could not load image. Error: {image_reader.errorString()}')

    # get md5
    md5sum = hashlib.md5(image_data).hexdigest()

    return image, md5sum


def jpeg_blob_from_image(image: QImage, quality: int = 100) -> bytes:
    if image.isNull():
        raise Exception('Image is null')

    blob = QByteArray()
    buffer = QBuffer(blob)
    buffer.open(QIODevice.WriteOnly)
    ok = image.save(buffer, 'jpg', quality)
    if not ok:
        raise Exception('Could not save image to buffer')

    return blob



from pathlib import Path
from random import choice
import shutil
from PySide6.QtCore import QUrl
from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import QFileDialog, QMessageBox, QPushButton, QWidget
from pamet.desktop_app.config import pamet_data_folder_path
from pamet import desktop_app
from pamet.desktop_app import get_user_settings

from pamet.model.script_note import ScriptNote
from pamet.views.note.card.props_edit.widget import FixedTextEdit
from .name_seed_data import adjectives, fruits, animals

from .ui_props_widget import Ui_ScriptNotePropsWidget

FILE_SCHEMA = 'file://'


def random_name():
    return f'{choice(adjectives)}-{choice(fruits)}-{choice(animals)}'


class ScriptNotePropsWidget(QWidget):

    def __init__(self, parent):
        QWidget.__init__(self, parent)

        self.ui = Ui_ScriptNotePropsWidget()
        self.ui.setupUi(self)

        self.edit_widget = parent

        # Locals
        self._unsaved_script_changes = False
        config = get_user_settings()
        self._scripts_folder = Path(config.scripts_folder)
        self._scripts_folder.mkdir(parents=True, exist_ok=True)

        # Add the note text edit
        self.text_edit = FixedTextEdit(self)
        self.ui.noteTextVerticalLayout.addWidget(self.text_edit)
        self.text_edit.textChanged.connect(self._handle_text_change)

        # Generate the create script buttons
        self._templates_folder = Path(config.script_templates_folder)
        for template in self._templates_folder.iterdir():
            button = QPushButton(template.stem)
            content = template.read_text()
            extension = template.suffix

            button.clicked.connect(lambda cnt=content, ext=extension: self.
                                   create_script_file(cnt, ext))
            self.ui.createScriptWidget.layout().addWidget(button)

        # Setup the UI according to the note props (is a script path set up)
        script_path = self.edit_widget.edited_note.script_path
        self.ui.scriptPathLineEdit.setText(str(script_path))
        if script_path is None:
            script_path = ''
        self.handle_script_path_change(str(script_path))

        self.ui.argumentsLineEdit.setText(self.edited_note.command_args)
        self.text_edit.setText(self.edited_note.text)

        # Connect the UI with the handlers
        self.ui.scriptPathLineEdit.textChanged.connect(
            self.handle_script_path_change)
        self.ui.scriptNameLineEdit.textChanged.connect(
            self.handle_script_name_change)
        self.ui.renamePushButton.clicked.connect(
            self.handle_rename_button_click)
        self.ui.openInEditorPushButton.clicked.connect(
            self._handle_open_in_editor_button_click)
        self.ui.openContainingFolderPushButton.clicked.connect(
            self._handle_open_containing_folder_button_click)
        self.ui.deletePushButton.clicked.connect(
            self._handle_delete_button_click)
        self.ui.chooseFilePushButton.clicked.connect(self._open_file_chooser)
        self.ui.argumentsLineEdit.textChanged.connect(
            self._handle_args_changed)
        self.ui.setDefaultScriptsFolder.clicked.connect(
            self._set_default_folder)
        self.ui.configureTemplatesButton.clicked.connect(
            self._configure_templates)
        self.ui.runPushButton.clicked.connect(
            lambda: desktop_app.script_runner.run(self.edited_note))
        self.ui.runInTerminalCheckbox.toggled.connect(
            self._handle_run_in_terminal_toggled)

    @property
    def edited_note(self) -> ScriptNote:
        return self.edit_widget.edited_note

    def create_script_file(self, content: str, extension: str):
        script_path = self._scripts_folder
        while script_path.exists():
            script_path = self._scripts_folder / random_name()
            script_path = script_path.with_suffix(extension)

        try:
            script_path.write_text(content)
            # script_path.chmod(script_path.stat().st_mode | stat.S_IEXEC)
        except Exception as e:
            QMessageBox(self, 'Could not write to file',
                        f'Could not create the script file. Exception: {e}')
            return

        self.ui.scriptPathLineEdit.setText(str(script_path))

    def _save_script_changes(self):
        script_path: Path = self.edited_note.script_path
        if not script_path:
            raise Exception

        script_path.write_text(self.edited_note.script_te)

    def _handle_text_change(self):
        self.edited_note.text = self.text_edit.toPlainText()

    def handle_script_path_change(self, new_path: str):
        old_path = self.edited_note.script_path

        if new_path.startswith(FILE_SCHEMA):
            new_path = new_path[len(FILE_SCHEMA):]
            self.ui.scriptPathLineEdit.setText(new_path)
            return

        new_path: Path = Path(new_path)
        self.edited_note.script_path = new_path

        if new_path.exists() and new_path.is_file():
            self.ui.pathErrorLabel.setText('Exists')
            self.ui.deletePushButton.setEnabled(True)
            self.ui.scriptNameLineEdit.setEnabled(True)
            self.ui.scriptNameLineEdit.setText(new_path.name)
            self.ui.openInEditorPushButton.setEnabled(True)
        else:
            if shutil.which(new_path):
                self.ui.pathErrorLabel.setText('Executable')
            else:
                self.ui.pathErrorLabel.setText('Missing')
            self.ui.deletePushButton.setEnabled(False)
            self.ui.scriptNameLineEdit.setEnabled(True)
            self.ui.scriptNameLineEdit.setText('')
            self.ui.openInEditorPushButton.setEnabled(False)

        note_text = self.edited_note.text
        # Keep the text matching the file name except if the user hasn't
        # made an edit (and they differ).
        if not note_text or (old_path and note_text == old_path.name):
            self.text_edit.setText(new_path.name)

    def handle_script_name_change(self, new_name: str):
        if new_name and new_name != self.edited_note.script_path.name:
            self.ui.renamePushButton.setEnabled(True)
        else:
            self.ui.renamePushButton.setEnabled(False)

    def handle_rename_button_click(self):
        script_path: Path = self.edited_note.script_path
        if not (script_path.exists() and script_path.is_file()):
            raise Exception(f'Cannot rename {script_path}. File missing.')

        new_name = self.ui.scriptNameLineEdit.text()
        new_path = script_path.parent / new_name
        try:
            # Make it possible to create a subdir on the spot
            new_path = new_path.resolve()
            new_path.parent.mkdir(parents=True, exist_ok=True)

            script_path.rename(new_path)
        except Exception as e:
            QMessageBox.warning(
                self, 'Error during rename',
                ('Could not rename the file {script_path} to {new_path}.'
                 f' Error: {e}'))
            return

        self.ui.scriptPathLineEdit.setText(str(new_path))

    def _handle_open_in_editor_button_click(self):
        QDesktopServices.openUrl(QUrl.fromLocalFile(str(self.edited_note.script_path)))

    def _handle_open_containing_folder_button_click(self):
        QDesktopServices.openUrl(QUrl.fromLocalFile(str(
            self.edited_note.script_path.parent)))

    def _handle_delete_button_click(self):
        path = self.edited_note.script_path
        try:
            path.unlink()
        except Exception as e:
            QMessageBox.warning(
                self, 'Error deleting file',
                f'Could not delete file {path}. Exception: {e}')
        self.ui.scriptPathLineEdit.setText('')

    def _open_file_chooser(self):
        path, _ = QFileDialog.getOpenFileName(self, 'Choose a script file',
                                              str(self._scripts_folder))
        if not path:
            return

        self.ui.scriptPathLineEdit.setText(path)

    def _handle_args_changed(self, new_args_text: str):
        self.edited_note.command_args = new_args_text

    def _set_default_folder(self):
        instructions_text = '''
        The config.json will be opened. You can change the scripts folder there.

        If you don't know what you're doing - it would be best to not touch it.
        '''
        QMessageBox.information(self, 'How to set the folder',
                                instructions_text)
        QDesktopServices.openUrl(QUrl.fromLocalFile(str(pamet_data_folder_path)))
        self.edit_widget._handle_esc_shortcut()

    def _configure_templates(self):
        instructions_text = '''
        You can add templates to the templates folder, which will be opened
        after you close this prompt.
        To restore the original templates - just delete the files and restart
        the app.
        '''
        QMessageBox.information(self, 'How to set the folder',
                                instructions_text)
        QDesktopServices.openUrl(QUrl.fromLocalFile(str(self._templates_folder)))
        self.edit_widget._handle_esc_shortcut()

    def _handle_run_in_terminal_toggled(self, checked: bool):
        self.edited_note.run_in_terminal = checked



import os
import shutil
from pathlib import Path
import subprocess
from PySide6.QtCore import QUrl
from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import QApplication, QMessageBox
import pamet
from pamet import desktop_app
from pamet.model.script_note import ScriptNote

python_missing_text = '''
Python is not installed on your system. Install it and try again.
It's best to do so with your package manager/ app store.
'''
dependency_missing_text = 'Dependency missing'


def google_url(search_text: str):
    url = f'https://www.google.com/search?q={search_text.replace(" ", "+")}'
    return url


def no_bash_warning():
    QMessageBox(
        QApplication.activeWindow(), 'No bash :(',
        ('Bash not found. Support for non-linux systems'
            ' is not fully implemented yet. Checkout the '
            'github issues for the project'))


class ScriptRunner:

    def __init__(self):
        pass

    def run(self, note: ScriptNote):
        # Veeery Ad-hoc ATM
        command = f'{note.script_path} {note.command_args}'
        path: Path = note.script_path

        if path.suffix == '.py':
            if os.name == 'nt':
                if not shutil.which('py'):
                    QMessageBox.information(QApplication.activeWindow(),
                                            dependency_missing_text,
                                            python_missing_text)

                    QDesktopServices.openUrl(
                        QUrl(google_url('microsoft store python')))
                    return
                command = f'py {command}'
            elif os.name == 'posix':
                if not shutil.which('python'):
                    QMessageBox.information(QApplication.activeWindow(),
                                            dependency_missing_text,
                                            python_missing_text)
                    QDesktopServices.openUrl(QUrl(
                        google_url('install python')))
                    return
                command = f'python {command}'
        elif path.suffix == '.sh':
            if os.name == 'nt':
                QMessageBox.warning(
                    QApplication.activeWindow(), 'Cannot run script',
                    ('Running bash scripts in Windows not implemented. '
                     'It\'s probably possible via WSL.'))
                return

            if not shutil.which('bash'):
                no_bash_warning()
                return
            command = f'bash {command}'

        config = desktop_app.get_user_settings()
        if note.run_in_terminal:
            if os.name == 'posix':
                if not shutil.which('bash'):
                    no_bash_warning()
                    return
                escaped = command.replace('"', '\\"')
                command = f'bash -c "{escaped}; bash"'
                command = f'{config.run_in_terminal_prefix_posix}{command}'
            elif os.name == 'nt':
                command = f'{config.run_in_terminal_prefix_windows}{command}'

        subprocess.run(command, shell=True)



from dataclasses import field
from datetime import datetime
from fusion import Entity, entity_type
from pamet.util.url import Url
from fusion.util import current_time, timestamp


@entity_type
class Page(Entity):
    name: str = ''
    created: str = field(default_factory=lambda: timestamp(current_time()))
    modified: str = field(default_factory=lambda: timestamp(current_time()))

    def __repr__(self):
        return f'<Page gid={self.gid()} name={self.name}>'

    def url(self):
        return Url(f'pamet:///p/{self.id}')

    @property
    def datetime_created(self) -> datetime:
        return datetime.fromisoformat(self.created)

    @datetime_created.setter
    def datetime_created(self, new_dt: datetime):
        self.created = timestamp(new_dt)

    @property
    def datetime_modified(self) -> datetime:
        return datetime.fromisoformat(self.modified)

    @datetime_modified.setter
    def datetime_modified(self, new_dt: datetime):
        self.modified = timestamp(new_dt)



# from PySide6.QtCore import Qt
# from .text_note_widget import TextNoteWidget
#
#
# class MarkdownNoteWidget(TextNoteWidget):
#     def __init__(self, note, parent=None):
#         super(MarkdownNoteWidget, self).__init__(note, parent=parent)
#
#         self.setTextFormat(Qt.MarkdownText)



from pathlib import Path
import time

# from PySide6.QtCore import QFileSystemWatcher
# from PySide6.QtWidgets import QApplication
from watchdog.observers import Observer
from watchdog.events import FileSystemEvent, FileSystemEventHandler


class FSWatchEventHandler(FileSystemEventHandler):
    def on_created(self, event):
        print('Created', event)

    def on_deleted(self, event):
        print('deleted', event)

    def on_modified(self, event):
        print('modified', event)

    def on_moved(self, event):
        print('moved', event)

    def on_any_event(self, event: FileSystemEvent):
        print('ANY:', event.event_type, event.src_path, event.is_directory)


class FSWatcherService:
    def __init__(self, folder: str | Path):
        self.folder = Path(folder).expanduser()
        if not self.folder.exists() or not self.folder.is_dir():
            raise Exception

        self.observer = Observer()
        event_handler = FSWatchEventHandler()
        observer = Observer()
        observer.schedule(event_handler, self.folder, recursive=False)
        observer.start()

    def start(self):
        self.observer.start()

    def stop(self):
        self.observer.stop()
        self.observer.join()

# class FSWatcherService(QFileSystemWatcher):
#     def __init__(self, folder: str | Path):
#         super().__init__()
#         self.folder = Path(folder).expanduser()
#         if not self.folder.exists() or not self.folder.is_dir():
#             raise Exception

#         self.addPath(str(self.folder))
#         self.fileChanged.connect(self.handle_file_changed)
#         self.directoryChanged.connect(self.handle_directory_changed)

#     def handle_file_changed(self, file_path):
#         file_path = Path(file_path)
#         print('file changed', file_path)

#     def handle_directory_changed(self, directory_path):
#         directory_path = Path(directory_path)
#         print('directory changed', directory_path)


# Test
if __name__ == '__main__':
    # app = QApplication()
    fs_watcher = FSWatcherService('~/fs_test')
    # app.exec()
    fs_watcher.start()
    try:
        while True:
            time.sleep(1)
    finally:
        fs_watcher.stop()



from typing import Callable, Union
from PySide6.QtWidgets import QLabel, QMenu, QWidgetAction
from PySide6.QtGui import QCursor, QKeySequence
from PySide6.QtCore import QPoint, Qt

import fusion
from fusion.libs.command import Command


def add_entries(menu: QMenu, entries):
    for name, command in entries.items():
        # Nested dicts mean submenus
        if isinstance(command, dict):
            submenu = QMenu(name, menu)
            add_entries(submenu, command)
            menu.addMenu(submenu)
        elif command is None:
            label_action = QWidgetAction(menu)
            label = QLabel(name)
            label.setAlignment(Qt.AlignCenter)
            label_action.setDefaultWidget(label)
            menu.addAction(label_action)
        elif isinstance(command, Callable):
            binding = None  # first_key_binding_for_command(command)
            if binding:
                menu.addAction(name, command, QKeySequence(binding.key))
            else:
                menu.addAction(name, command)
        else:
            raise Exception


class ContextMenuWidget(QMenu):

    def __init__(self, parent: str, entries: dict[str, Union[Command, dict]]):
        QMenu.__init__(self, parent)

        add_entries(self, entries)

        self.aboutToHide.connect(self.hiding)

        # This should probably be done via a state update
        fusion.call_delayed(self.popup_on_mouse_pos, 0)

    def popup_on_mouse_pos(self):
        position = QCursor.pos()
        self.popup(position)

    def hiding(self):
        self.close()

    def close(self):
        self.deleteLater()



from pathlib import Path
from typing import Tuple
import numpy as np
from fusion.libs.channel import Channel

from pamet.services.search.base import BaseSearchService, IndexEntry, SearchResult

try:
    from sqlitedict import SqliteDict
    from sentence_transformers import SentenceTransformer, util
except Exception:
    raise Exception('The semantic search service depends on the python '
                    'packages ')


class SemanticSearchService(BaseSearchService):

    def __init__(self, data_folder: str, change_set_channel: Channel):
        super().__init__(change_set_channel)
        self.data_folder = Path(data_folder)
        self.data_folder.mkdir(parents=True, exist_ok=True)
        self.fv_index = {}  # Feature vectors by text
        self.sqlite_db = SqliteDict(self.data_folder / 'semantic_index.sqlite',
                                    autocommit=True)
        self.embedding_model = SentenceTransformer('all-MiniLM-L6-v2')
        self.report_item_count = False

    def upsert_to_index(self, entry: IndexEntry):
        if entry.content_string not in self.fv_index:
            self._get_fv(entry.content_string)
        super().upsert_to_index(entry)
        if self.report_item_count:
            print(f'Items in index: {len(self.fv_index)}', end='\r')

    def remove_from_index(self, note_gid: Tuple):
        self.fv_index.pop(note_gid)
        super().remove_from_index(note_gid)
        if self.report_item_count:
            print(f'Items in index: {len(self.fv_index)}', end='\r')

    def _get_fv(self, text: str) -> dict:
        """Returns a feature vector for the given text either from the cache
        or by calculating it"""
        fv = self.fv_index.get(text)

        if fv is None:
            fv = self.embedding_model.encode([text],
                                             show_progress_bar=False)[0]
            self.sqlite_db[text] = fv
            self.fv_index[text] = fv
        return fv

    def text_search(self, text: str) -> SearchResult:
        search_results = []
        text = text.lower()
        query_fv = self.embedding_model.encode([text],
                                               show_progress_bar=False)[0]
        for note_gid, index_entry in self.index.items():
            score = np.linalg.norm(self.fv_index[index_entry.content_string] -
                                   query_fv)

            if score < 10:
                search_results.append(
                    SearchResult(
                        note_gid=note_gid,
                        content_string=(f'{score:.2f}'
                                        f' {index_entry.content_string}'),
                        score=score))

        search_results = sorted(search_results, key=lambda r: r.score)
        return search_results

    def load_all_content(self):
        # Load the feature vectors from the database
        print('Loading feature vectors from database...', end='')
        for text, fv in self.sqlite_db.items():
            self.fv_index[text] = fv
        print('Done')

        self.report_item_count = True
        super().load_all_content()
        self.report_item_count = False



from PySide6.QtGui import QShortcut, QKeySequence
from PySide6.QtWidgets import QLabel, QWidget
from PySide6.QtCore import Qt

import fusion
from fusion.util.point2d import Point2D
from fusion.libs.entity.change import Change
from fusion.platform.qt_widgets import bind_and_apply_state
from fusion.view import View

from pamet import actions, note_view_type_by_state
from pamet.actions import tab as tab_actions
from pamet.model.text_note import TextNote
from pamet.views.map_page.properties_widget import MapPagePropertiesWidget

from pamet.views.map_page.widget import MapPageWidget
from pamet.views.search_bar.widget import SearchBarWidget, SearchBarWidgetState
from pamet.views.semantic_search_bar.widget import SemanticSearchBarWidget, SemanticSearchBarWidgetState
from pamet.views.tab.ui_widget import Ui_TabMainWidget

from pamet.views.map_page.state import MapPageViewState

SIDEBAR_WIDTH = 300
MIN_TAB_WIDTH = SIDEBAR_WIDTH * 1.1
MAX_PAGE_CACHE_SIZE = 100


class TabWidget(QWidget, View):

    def __init__(self, parent, initial_state):
        QWidget.__init__(self, parent=parent)
        View.__init__(self, initial_state=initial_state)

        # That's because of a Qt bug - when opening a new tab - the central
        # widget of the tab has a QStackWidget as parent which does not have
        # the window as a parent
        self.parent_window = parent

        self.ui = Ui_TabMainWidget()
        self.ui.setupUi(self)

        self._page_view = None
        self._page_subscription_id = None
        self.edit_view = None
        self._left_sidebar_widget = None
        self._right_sidebar_widget = None
        self._search_widget = None
        self._semantic_search_widget = None
        self._page_widget_cache = {}  # By page_id

        new_note_shortcut = QShortcut(QKeySequence('N'), self)
        new_note_shortcut.activated.connect(self.create_new_note_command)

        new_page_shortcut = QShortcut(QKeySequence('ctrl+N'), self)
        new_page_shortcut.activated.connect(self.create_new_page_command)

        self.ui.rightSidebarCloseButton.clicked.connect(
            lambda: tab_actions.close_page_properties(self.state()))
        self.ui.leftSidebarCloseButton.clicked.connect(
            lambda: tab_actions.close_left_sidebar(self.state()))

        self.ui.commandLineEdit.hide()
        self.ui.leftSidebarContainer.hide()

        self.page_not_found_label = QLabel('Page missing.', self)

        # The state should be managed by the window
        bind_and_apply_state(self,
                             initial_state,
                             on_state_change=self.on_state_change)

    def page_view(self) -> MapPageViewState:
        return self._page_view

    def set_search_pane_state(self, pane_state: bool):
        if pane_state:
            if type(pane_state) == SearchBarWidgetState:
                if not self._search_widget:
                    self._search_widget = SearchBarWidget(
                        self, initial_state=pane_state)

                self._left_sidebar_widget = self._search_widget
                self.ui.leftSidebarContainer.layout().addWidget(
                    self._search_widget)

            elif type(pane_state) == SemanticSearchBarWidgetState:
                if not self._semantic_search_widget:
                    self._semantic_search_widget = SemanticSearchBarWidget(
                        self, initial_state=pane_state)

                self._left_sidebar_widget = self._semantic_search_widget
                self.ui.leftSidebarContainer.layout().addWidget(
                    self._semantic_search_widget)

            self.ui.leftSidebarContainer.show()
            self.ui.leftSidebarContainer.raise_()
        else:
            self.ui.leftSidebarContainer.hide()

    def on_state_change(self, change: Change):
        state = change.last_state()
        if change.updated.page_view_state or not self.page_view():
            self.switch_to_page_view(state.page_view_state)

        if change.updated.title:
            self_idx = self.parent_window.ui.tabBarWidget.indexOf(self)
            self.parent_window.ui.tabBarWidget.setTabText(
                self_idx, state.title)

        if change.updated.page_edit_view_state:
            if self._right_sidebar_widget:
                self.ui.rightSidebarContainer.layout().removeWidget(
                    self._right_sidebar_widget)
                self._right_sidebar_widget.deleteLater()
                self._right_sidebar_widget = None

            if state.page_edit_view_state:
                new_widget = MapPagePropertiesWidget(
                    self, initial_state=state.page_edit_view_state)
                self._right_sidebar_widget = new_widget
                self.ui.rightSidebarContainer.layout().addWidget(new_widget)
                self.ui.rightSidebarContainer.show()
                self.ui.rightSidebarContainer.raise_()
            else:
                self.ui.rightSidebarContainer.hide()

        if change.updated.left_sidebar_state:
            # Clear the sidebar
            if self._left_sidebar_widget:
                self.ui.leftSidebarContainer.layout().removeWidget(
                    self._left_sidebar_widget)

                # If it's the search bar - don't delete it, we'll reuse it
                if self._left_sidebar_widget not in [
                        self._search_widget, self._semantic_search_widget
                ]:
                    self._left_sidebar_widget.deleteLater()

                self._left_sidebar_widget = None

            self.set_search_pane_state(state.left_sidebar_state)

        if change.updated.note_edit_view_state:
            if self.edit_view:
                self.edit_view.close()
                self.edit_view.deleteLater()
                self.edit_view = None

            if state.note_edit_view_state:
                EditViewType = note_view_type_by_state(
                    type(state.note_edit_view_state).__name__)
                self.edit_view = EditViewType(
                    parent=self, initial_state=state.note_edit_view_state)
                self.edit_view.setParent(self)
                self.edit_view.setWindowFlag(Qt.Sheet, True)
                self.edit_view.show()
                self.edit_view.setFocus()

    def add_page_wideget_to_cache(self, page_widget: MapPageWidget):
        self._page_widget_cache[page_widget.state().view_id] = page_widget
        if MAX_PAGE_CACHE_SIZE < len(self._page_widget_cache):
            page_id, page_widget = self._page_widget_cache.popitem()
            page_widget.deleteLater()

    def cached_page_widget(self, page_view_id: str):
        return self._page_widget_cache.pop(page_view_id, None)

    def switch_to_page_view(self, page_view_state: MapPageViewState):
        """Switches to the specified page view state.

        The TabWidget keeps a cache with page widgets associated with their
        state ids. When possible - those get reused.
        """
        if self._page_view:
            self.ui.centralContainer.layout().removeWidget(self._page_view)
            if isinstance(self._page_view, MapPageWidget):
                self.add_page_wideget_to_cache(self._page_view)

            self._page_view.hide()

        if not page_view_state:
            page_widget = self.page_not_found_label
        else:
            page_widget = self.cached_page_widget(page_view_state.view_id)
            if not page_widget:
                page_widget = MapPageWidget(parent=self,
                                            initial_state=page_view_state)
        self._page_view = page_widget

        self.ui.centralContainer.layout().addWidget(self._page_view)
        self._page_view.show()

        # Must be visible to accept focus, so queue on the main loop
        fusion.call_delayed(self._page_view.setFocus)

    def create_new_note_command(self):
        page_widget = self.page_view()

        # If the mouse is on the page - make the note on its position.
        # Else: make it in the center of the viewport
        if page_widget.underMouse():
            edit_window_pos = self.cursor().pos()
        else:
            edit_window_pos = page_widget.geometry().center()

        edit_window_pos = Point2D(edit_window_pos.x(), edit_window_pos.y())
        note_pos = page_widget.state().unproject_point(edit_window_pos)
        new_note = TextNote()
        new_note = new_note.with_id(page_id=page_widget.state().page_id)
        new_note.x = note_pos.x()
        new_note.y = note_pos.y()
        actions.note.create_new_note(self.state(), new_note)

    def create_new_page_command(self):
        mouse_pos = self.cursor().pos()
        edit_window_pos = Point2D(mouse_pos.x(), mouse_pos.y())
        tab_actions.create_new_page(self.state(), edit_window_pos)



from pamet.services.search.base import BaseSearchService, SearchResult


class PlainSearchService(BaseSearchService):
    def text_search(self, text: str) -> SearchResult:
        """Searches the cached texts and returns search results"""
        search_results = []
        text = text.lower()
        for note_gid, index_entry in self.index.items():
            score = 0
            if index_entry.content_lowered.startswith(text):
                score = 1
            elif index_entry.content_lowered.find(text):
                score = 0.99

            if score <= 0:
                continue

            search_result = SearchResult(note_gid)
            search_results.append(search_result)

        return sorted(search_results, key=lambda r: r.score, reverse=True)



from __future__ import annotations

import fusion
from fusion.libs.entity import dump_to_dict
from fusion.util import current_time
from fusion import fsm
from fusion.libs.action import action

import pamet
from pamet.util import snap_to_grid
from pamet.model.note import Note
from pamet.views.note.qt_helpers import minimal_nonelided_size
from pamet.actions import map_page as map_page_actions

log = fusion.get_logger(__name__)


@action('notes.create_new_note')
def create_new_note(tab_state: TabViewState, note: Note):
    # Check if there's an open edit window and abort it if so
    if tab_state.note_edit_view_state:
        abort_editing_note(tab_state)

    EditViewType = pamet.note_view_state_type_for_note(note, edit=True)
    edit_view_state = EditViewType(id=note.gid(),
                                   new_note_dict=dump_to_dict(note))
    edit_view_state.update_from_note(note)

    tab_state.note_edit_view_state = edit_view_state

    fsm.add_state(edit_view_state)
    fsm.update_state(tab_state)


@action('notes.finish_creating_note')
def finish_creating_note(tab_state: TabViewState, note: Note):
    # note = pamet.create_note(**note.asdict())

    note.datetime_created = current_time()
    note.datetime_modified = current_time()

    # Autosize the note
    rect = note.rect()
    rect.set_size(snap_to_grid(minimal_nonelided_size(note)))
    note.set_rect(rect)
    pamet.insert_note(note)

    tab_state.note_edit_view_state = None
    fsm.update_state(tab_state)


@action('notes.start_editing_note')
def start_editing_note(tab_view_state: TabViewState, note: Note):
    # Check if there's an open edit window and abort it if so
    if tab_view_state.note_edit_view_state:
        abort_editing_note(tab_view_state)

    EditViewType = pamet.note_view_state_type_for_note(note, edit=True)
    edit_view_state = EditViewType(id=note.gid())
    edit_view_state.update_from_note(note)
    tab_view_state.note_edit_view_state = edit_view_state

    fsm.add_state(edit_view_state)
    fsm.update_state(tab_view_state)


@action('notes.finish_editing_note')
def finish_editing_note(tab_state: TabViewState, note: Note):
    pamet.update_note(note)
    fsm.remove_state(tab_state.note_edit_view_state)
    tab_state.note_edit_view_state = None
    fsm.update_state(tab_state)
    # autosize_note(note_component_id)


@action('notes.abort_editing_note')
def abort_editing_note(tab_state: TabViewState):
    tab_state.note_edit_view_state = None
    fsm.update_state(tab_state)


@action('notes.abort_editing_and_delete_note')
def abort_editing_and_delete_note(tab_state: TabViewState):
    note = tab_state.note_edit_view_state.get_note()
    map_page_actions.delete_notes_and_connected_arrows([note])
    abort_editing_note(tab_state)


@action('note.apply_page_change_to_anchor_view')
def apply_page_change_to_anchor_view(
        note_view_state: tab_widget.TextNoteViewState):
    page = pamet.page(note_view_state.url.get_page_id())
    # If it's missing - leave the text, the border will indicate
    if page:
        note_view_state.text = page.name
    else:
        note_view_state.text += '(removed)'
    fsm.update_state(note_view_state)


@action('note.switch_note_type')
def switch_note_type(tab_state: TabViewState, note: Note):
    if tab_state.note_edit_view_state.create_mode:
        abort_editing_note(tab_state)
        create_new_note(tab_state, note)
    else:
        finish_editing_note(tab_state, note)
        start_editing_note(tab_state, note)



from copy import copy
import pytest

from fusion.storage.in_memory_repository import InMemoryRepository
import pamet
from pamet.model.note import Note
from pamet.model.page import Page
from pamet.model.text_note import TextNote
from pamet.storage.file_system.repository import FSStorageRepository


def test_in_memory_repo():
    repo = InMemoryRepository()

    page1 = Page()
    page2 = Page()

    note1a = Note.in_page(page1)
    note1b_for_update = TextNote.in_page(page1)
    note1c_for_removal = Note.in_page(page1)

    note2 = Note.in_page(page2)

    # Test the update and remove exceptions for missing entities
    with pytest.raises(Exception):
        repo.update_one(page1)

    with pytest.raises(Exception):
        repo.remove_one(page1)

    entities = [
        page1, page2, note1a, note1b_for_update, note1c_for_removal, note2
    ]
    # Add all entities
    for entity in entities:
        repo.insert_one(entity)

    # Test note immutability
    with pytest.raises(Exception):
        note1b_for_update.text = 'test'

    note1b_for_update = note1b_for_update.copy()
    note1b_for_update.text = 'test'
    repo.update_one(note1b_for_update)
    repo.remove_one(note1c_for_removal)

    expected_entities = copy(entities)
    expected_entities.remove(note1c_for_removal)

    # Test find operations

    # Find all
    assert set(repo.find()) == set(expected_entities)

    # Find using the global id index
    assert repo.find_one(gid=note1b_for_update.gid()).asdict() == \
        note1b_for_update.asdict()

    # Find using the parent_gid index
    assert set(repo.find(parent_gid=page1.gid())) == \
        set([note1a, note1b_for_update])

    # Find using the type name index
    assert set(repo.find(type=TextNote)) == \
        set([note1b_for_update])


def test_fs_repo_CRUD(tmp_path):
    fs_repo = FSStorageRepository.new(tmp_path)
    pamet.set_sync_repo(fs_repo)

    page = Page(id='test_page')
    pamet.insert_page(page)
    fs_repo.write_to_disk()

    assert fs_repo.path_for_page(page).exists()

    note = TextNote()
    note.text = 'test text'
    pamet.insert_note(note, page)
    fs_repo.write_to_disk()

    page_json = FSStorageRepository.serialize_page(page, pamet.notes(page),
                                                   pamet.arrows(page))

    assert fs_repo.path_for_page(page).read_text() == page_json

    pamet.remove_page(page)
    fs_repo.write_to_disk()

    assert list(pamet.pages()) == []



from copy import copy
import pamet
from pamet.model.page import Page
from pamet.model.text_note import TextNote
from pamet.storage.pamet_in_memory_repo import PametInMemoryRepository


def test_page_CRUD():
    fs_repo = PametInMemoryRepository()
    pamet.set_sync_repo(fs_repo)

    page = Page(name='test_page')
    pamet.insert_page(copy(page))
    assert [p.asdict() for p in pamet.pages()] == [page.asdict()]

    page.name = 'test_page_updated'
    pamet.update_page(page)
    assert [p.asdict() for p in pamet.pages()] == [page.asdict()]

    pamet.remove_page(page)
    assert list(pamet.pages()) == []


def test_note_CRUD():
    fs_repo = PametInMemoryRepository()
    pamet.set_sync_repo(fs_repo)

    page = Page(name='test_page')
    pamet.insert_page(page)

    note = TextNote.in_page(page)
    note.text = 'test text'
    pamet.insert_note(copy(note))
    assert [n.asdict() for n in pamet.notes(page)] == [note.asdict()]

    note.text = 'test text changed'
    pamet.update_note(note)
    assert pamet.find_one(gid=note.gid()).asdict() == note.asdict()

    pamet.remove_note(note)
    assert pamet.find_one(gid=note.gid()) is None



#!/usr/bin/env python

import argparse
import os
from pathlib import Path
import shutil

LOCAL = os.path.expanduser('~/.local')
LOCAL_BIN = os.path.join(LOCAL, 'bin')
LOCAL_LIB = os.path.join(LOCAL, 'lib')
LOCAL_APPS = os.path.join(LOCAL, 'share', 'applications')
ICONS_DIR = os.path.join(LOCAL, 'share', 'icons', 'hicolor', '256x256', 'apps')
SOURCE_PATH = Path(__file__).parent
PAMET_RESOURCES_PATH = SOURCE_PATH / 'pamet' / 'resources'

for path in [LOCAL, LOCAL_APPS, LOCAL_BIN, LOCAL_LIB, ICONS_DIR]:
    os.makedirs(path, exist_ok=True)


def main():
    parser = argparse.ArgumentParser()
    # parser.add_argument('build_dir', help='Location of the release build')
    parser.add_argument('--uninstall', action='store_true')

    args = parser.parse_args()
    # build_dir = args.build_dir
    uninstall = args.uninstall

    # if not os.path.exists(build_dir) or not os.path.isdir(build_dir):
    #     print('Invalid build dir')

    def apply(build_path, install_path):
        if uninstall:
            if not os.path.exists(install_path):
                print('Missing file at install path', install_path)
                return

            print('Removing', install_path)
            os.remove(install_path)
        else:
            if os.path.lexists(install_path):
                print('Removing existing file at install path', install_path)
                os.remove(install_path)

            print('Copying %s to %s' % (build_path, install_path))
            shutil.copy(build_path, install_path)

    # Icon
    icon_path = PAMET_RESOURCES_PATH / 'icons' / 'pamet-icon.png'
    icon_install_path = os.path.join(ICONS_DIR, 'pamet.png')
    apply(icon_path, icon_install_path)

    # Desktop entry
    desktop_entry = PAMET_RESOURCES_PATH / 'desktop_entries' / 'pamet.desktop'
    apply(desktop_entry, os.path.join(LOCAL_APPS, 'pamet.desktop'))

    os.system('update-desktop-database ~/.local/share/applications')


if __name__ == '__main__':
    main()



from pathlib import Path

from PySide6.QtCore import QByteArray, QUrl, Qt
from PySide6.QtGui import QImage, QPixmap, QResizeEvent
from PySide6.QtNetwork import QNetworkReply, QNetworkRequest
from PySide6.QtWidgets import QFileDialog, QWidget
from fusion.util.point2d import Point2D
from pamet.desktop_app.util import get_image_and_md5_from_bytearray

from pamet.util.url import Url
from pamet.desktop_app import media_store
from pamet.model.card_note import CardNote
from pamet.model.image_note import ImageNote

from .ui_widget import Ui_ImagePropsWidget

MAX_IMAGE_SIZE_EDGE = 2048
MAX_IMAGE_SIZE_BYTES = 30 * 1000 * 1000


class ImagePropsWidget(QWidget):

    def __init__(self, parent):
        super().__init__(parent)

        self.edit_widget = parent

        # Locals
        self._image_preview = None

        self.ui = Ui_ImagePropsWidget()
        self.ui.setupUi(self)

        # Setup the image props widget
        if isinstance(self.edit_widget.edited_note, (ImageNote, CardNote)):
            # Set the url line edit text and then connect the handler (to avoid
            # triggering it beforehand)
            self.ui.urlLineEdit.setText(
                str(self.edit_widget.edited_note.image_url))

            # Setup the image from the local data if any
            local_url: Url = self.edit_widget.edited_note.local_image_url
            if local_url:
                path = local_url.path
                if local_url.is_internal():
                    path = media_store().path_for_internal_uri(local_url)

                image = QImage(path)
                if image.isNull():
                    self.update_image_note_and_preview(
                        error=f'Could not load local image at "{path}"')
                else:
                    self.update_image_note_and_preview(local_url, image)

            self.ui.urlLineEdit.textChanged.connect(
                self.handle_image_url_text_change)

            # If there's been a problem and the local url is missing - trigger
            # a download
            if self.edit_widget.edited_note.image_url and \
                    not self.edit_widget.edited_note.local_image_url:
                self.handle_image_url_text_change(
                    self.edit_widget.edited_note.image_url)

        # Image props connect
        self.ui.urlLineEdit.textChanged.connect(
            self.handle_image_url_text_change)
        self.ui.openFileButton.clicked.connect(
            self.handle_open_file_button_click)

    @property
    def _image_download_reply(self) -> QNetworkReply:
        return self.edit_widget._image_download_reply

    @_image_download_reply.setter
    def _image_download_reply(self, reply: QNetworkReply):
        self.edit_widget._image_download_reply = reply

    def handle_open_file_button_click(self):
        file_path, _ = QFileDialog.getOpenFileName()
        self.ui.urlLineEdit.setText(file_path)

    def resizeEvent(self, event: QResizeEvent) -> None:
        if self._image_preview and \
                self.edit_widget.edited_note.local_image_url:
            self.update_image_note_and_preview(
                local_image_url=self.edit_widget.edited_note.local_image_url,
                image=self._image_preview)
        return super().resizeEvent(event)

    def update_image_note_and_preview(self,
                                      local_image_url: Url = None,
                                      image: QImage = None,
                                      md5sum: str = None,
                                      error: str = None):
        note = self.edit_widget.edited_note
        # Set the note image_url
        url_input = self.ui.urlLineEdit.text()
        note.image_url = url_input

        if error:
            if local_image_url or image or md5sum:
                raise Exception
            note.local_image_url = None
            self.ui.infoLabel.setText(error)
            self._image_preview = None
            return

        self._image_preview = image
        note.local_image_url = local_image_url
        note.image_size = Point2D(image.size().width(), image.size().height())
        note.image_md5 = md5sum

        w = self.ui.infoLabel.size().width()
        h = self.ui.infoLabel.size().height()
        pixmap = QPixmap.fromImage(image)
        pixmap = pixmap.scaled(w, h, Qt.KeepAspectRatio)
        self.ui.infoLabel.setPixmap(pixmap)

    def handle_image_url_text_change(self, new_text: str):
        if not new_text:
            self.update_image_note_and_preview(error='Image not specified')
            return

        url = Url(new_text)
        if url.scheme == 'file' or not url.scheme and Path(str(url)).exists():
            # Check if exists - if not - error
            url_path = Path(url.path)
            if not url_path.exists():
                self.update_image_note_and_preview(error='File does not exist')
                return
            # Check size before reading - if too big - error (for now)
            elif url_path.stat().st_size > MAX_IMAGE_SIZE_BYTES:
                self.update_image_note_and_preview(error='Image is too big')
                return

            # Read binary
            with open(url_path, 'rb') as image_file:
                image_data = image_file.read()

            # Convert to QImage
            try:
                image, md5sum = get_image_and_md5_from_bytearray(
                    QByteArray(image_data))
            except Exception as e:
                self.update_image_note_and_preview(
                    error=f'Could not load image: {e}')
                return

            self.update_image_note_and_preview(local_image_url=url.path,
                                               image=image,
                                               md5sum=md5sum)
        elif url.is_internal():
            path = media_store().path_for_internal_uri(url)
            image = QImage(path)
            self.update_image_note_and_preview(path, image)
        elif url.has_web_schema():
            # Check if already downloaded - no. No such optimisations
            # (url <> path associations) present

            # Cancel previous download if any
            if self._image_download_reply:
                self._image_download_reply.abort()

            # Start downloading
            self._image_download_reply = self.edit_widget._network_am.get(
                QNetworkRequest(QUrl(str(url))))

            # Update label to mark downloading
            self.update_image_note_and_preview(error='Downloading...')

            # Done handler: remove reply, preview, save
            def on_complete():
                # This handler is called after on_error if one occurres
                if self._image_download_reply.error():
                    self.update_image_note_and_preview(
                        error=(f'Could not download image. Error:'
                               f' {self._image_download_reply.errorString()}'))
                    self._image_download_reply = None
                    return

                image_qbytearray = self._image_download_reply.readAll()
                image_data = image_qbytearray.data()
                if not image_data:
                    self.update_image_note_and_preview(error=(
                        'Download complete, but no image data to read.'))
                    return

                try:
                    image, md5sum = get_image_and_md5_from_bytearray(
                        image_qbytearray)
                except Exception as e:
                    self.update_image_note_and_preview(
                        error=f'Could not read image. Error: {e}')
                    return

                # Check image size in bytes and scale down if needed
                if image.sizeInBytes() > MAX_IMAGE_SIZE_BYTES:
                    self.update_image_note_and_preview(
                        error='Image too big. Resizing before save.')
                    image = image.scaled(MAX_IMAGE_SIZE_EDGE,
                                         MAX_IMAGE_SIZE_EDGE,
                                         Qt.KeepAspectRatio)

                try:
                    local_url = media_store().save_image(
                        image, 'jpg', self.edit_widget.state().page_id, url)
                except Exception as e:
                    self.update_image_note_and_preview(
                        error=f'Could not save image to media store: {e}')
                else:

                    self.update_image_note_and_preview(
                        local_image_url=local_url, image=image, md5sum=md5sum)
                self._image_download_reply = None

            # def on_error(error_code: QNetworkReply.NetworkError):

            def on_progress(bytes_received: int, bytes_total: int):
                if bytes_total <= 0:
                    progress = f'{bytes_received / 1000:.2f}k/??'
                else:
                    progress = f'{bytes_received / bytes_total * 100:.2f}%'

                self.update_image_note_and_preview(
                    error=f'Download at {progress}')

            # Connect the handlers
            self._image_download_reply.finished.connect(on_complete)
            # self._image_download_reply.errorOccurred.connect(on_error)
            self._image_download_reply.downloadProgress.connect(on_progress)

        else:
            self.ui.infoLabel.setText('Url schema not recognized')



from PySide6.QtCore import QSize, QUrl
from PySide6.QtNetwork import QNetworkReply, QNetworkRequest
from PySide6.QtWidgets import QTextEdit, QWidget

import pamet
from pamet.util.url import Url
from .ui_widget import Ui_TextEditPropsWidget

TEXT_EDIT_MIN_SIZE = 30


class FixedTextEdit(QTextEdit):
    """The default QTextEdit implementation has an inflexible (borderline
    buggy) size constraints, which cannot be overwriteen in any other way."""

    def minimumSizeHint(self) -> QSize:
        return QSize(TEXT_EDIT_MIN_SIZE, TEXT_EDIT_MIN_SIZE)

    def sizeHint(self) -> QSize:
        return QSize(TEXT_EDIT_MIN_SIZE, TEXT_EDIT_MIN_SIZE)


class TextEditPropsWidget(QWidget):

    def __init__(self, parent):
        super().__init__(parent)
        self.edit_widget = parent

        self.ui = Ui_TextEditPropsWidget()
        self.ui.setupUi(self)
        layout = self.layout()
        self.text_edit = FixedTextEdit(self)
        layout.addWidget(self.text_edit)
        self.setLayout(layout)

        # Connect stuff
        self.ui.getTitleButton.clicked.connect(self._get_title)
        self.text_edit.textChanged.connect(self.on_text_change)

    @property
    def _image_download_reply(self) -> QNetworkReply:
        return self.edit_widget._image_download_reply

    @_image_download_reply.setter
    def _image_download_reply(self, reply: QNetworkReply):
        self.edit_widget._image_download_reply = reply

    def on_text_change(self):
        new_text = self.text_edit.toPlainText().strip()
        self.edit_widget.edited_note.text = new_text

    def _get_title(self):
        edited_note = self.edit_widget.edited_note
        url: Url = edited_note.url
        page = pamet.page(url.get_page_id())

        if page:
            self.text_edit.setText(page.name)
        elif url.is_external():
            if not url.scheme:
                url_str = 'https://' + str(url)
            else:
                url_str = str(url)
            self._image_download_reply = self.edit_widget._network_am.get(
                QNetworkRequest(QUrl(url_str)))

            self.ui.downloadInfoLabel.show()

            # Done handler: remove reply, preview, save
            def on_complete():
                # This handler is called after on_error if one occurres
                if self._image_download_reply.error():
                    self.ui.downloadInfoLabel.setText('Download error')
                    return
                html = self._image_download_reply.readAll().toStdString()
                title = html[html.find('<title>') + 7:html.find('</title>')]
                self.text_edit.setText(title)
                self.ui.downloadInfoLabel.hide()

            def on_progress(bytes_received: int, bytes_total: int):
                if bytes_total <= 0:
                    progress = f'{bytes_received / 1000:.2f}k/??'
                else:
                    progress = f'{bytes_received / bytes_total * 100:.2f}%'

                self.ui.downloadInfoLabel.setText(f'Download at {progress}')

            # Connect the handlers
            self._image_download_reply.finished.connect(on_complete)
            self._image_download_reply.downloadProgress.connect(on_progress)



from dataclasses import field

import pamet
from fusion.libs.state import view_state_type, ViewState
from pamet.model.note import Note


@view_state_type
class NoteViewState(Note, ViewState):
    badges: list = field(default_factory=list, init=False, repr=False)

    @property
    def note_gid(self):
        return self.id

    def __post_init__(self):
        if not self.note_gid:
            raise Exception('All note views should have a mapped note.')

    def update_from_note(self, note: Note):
        note_props = note.asdict()
        note_props.pop('id')
        self.replace(**note_props)

    def get_note(self):
        return pamet.find_one(gid=self.note_gid)



from __future__ import annotations
from importlib import resources
import importlib.metadata

__version__ = importlib.metadata.version(__package__)

from typing import Generator, List, Union

import fusion
from fusion.change_aggregator import ChangeAggregator
from fusion.libs.entity.change import Change
from fusion.libs.channel import Channel
from fusion.storage.repository import Repository
from fusion.libs.entity import Entity
from fusion.extensions_loader import ExtensionsLoader
from fusion.libs.action import completed_root_actions
from fusion.logging import get_logger

from pamet.storage.base_repository import PametRepository
from pamet.storage.pamet_in_memory_repo import PametInMemoryRepository
from pamet.services.clipboard import ClipboardService
from pamet.services.search.base import BaseSearchService
from pamet import channels
from pamet.model.arrow import Arrow
from pamet.desktop_app import get_user_settings

from .note_view_lib import register_note_view_type, note_view_type
from .note_view_lib import note_view_type_by_state, note_state_type_by_view
from .note_view_lib import note_view_state_type_for_note

log = get_logger(__name__)

model_dir = resources.files('pamet') / 'model'
pamet_root = model_dir.parent
entity_types_loader = ExtensionsLoader(pamet_root)
entity_types_loader.load_all_recursively(model_dir)

clipboard = ClipboardService()

_search_service = None
_semantic_search_service = None
_broken_entities = {}  # the entity as key, and exception as value

_sync_repo: PametRepository = None  # Should be set via the setter method

# _persistence_manager = PersistenceManager()
_undo_service = None

raw_entity_changes = Channel('__RAW_ENTITY_CHANGES__')

entity_change_aggregator = None


# Chain the entity changes channels
def unwrap_changeset(changeset: List[Change]):
    for change in changeset:
        channels.entity_changes_by_id.push(change)
        channels.entity_changes_by_parent_gid.push(change)


def setup():
    global entity_change_aggregator
    global clipboard

    entity_change_aggregator = ChangeAggregator(
        input_channel=raw_entity_changes,
        release_trigger_channel=completed_root_actions,
        changeset_output_channel=channels.entity_change_sets_per_TLA)
    clipboard = ClipboardService()

    channels.entity_change_sets_per_TLA.subscribe(unwrap_changeset)


setup()


def reset():
    setup()


def search_service() -> BaseSearchService:
    return _search_service


def set_search_service(service):
    global _search_service
    _search_service = service


def semantic_search_service() -> BaseSearchService:
    return _semantic_search_service


def set_semantic_search_service(service):
    global _semantic_search_service
    _semantic_search_service = service


def broken_entities():
    return _broken_entities


def entity_broke(entity: Entity, exception: Exception):
    if entity not in _broken_entities:
        log.error(f'Entity {entity} broke.')

    _broken_entities[entity] = exception


def undo_service():
    return _undo_service


def set_undo_service(undo_service_):
    global _undo_service
    _undo_service = undo_service_


def sync_repo():
    """Returns the synchronous repo"""
    if not _sync_repo:
        raise Exception(
            '''Sync repo not set. To use pamet, you should configure
            a repository first:
            # Instantiate a repository (e.g. a file-system based one)
            repo = FSStorageRepository(folder_path)
            # Set it as the sync repo
            set_sync_repo(repo)
            ''')

    return _sync_repo


def set_sync_repo(repository: Repository):
    """
    Set a repository, which will be used synchronously when calling for CRUD
    operations.
    """
    global _sync_repo
    repository.set_change_channel(raw_entity_changes)
    _sync_repo = repository


set_sync_repo(PametInMemoryRepository())


@log.traced
def async_repo():
    """Returns the asynchronous repository"""
    return _persistence_manager.async_repo


@log.traced
def set_async_repo(repo: Repository):
    """Set the secondary repository, that's called asynchronously to persist
    the changes.

    This is used for the GUI interfaces, where the changes are
    saved synchronously in the in-memory repo initially to avoid UI lag. And
    then pushed to the async repo too. The errors or delays from the async repo
    are handled by the PersistenceManager"""

    log.info('Setting repo to %s' % repo)
    _persistence_manager.async_repo = repo

    # Here the persistence manager should be connected to the entityTLA channel


# ------------Finds-----------------
def find(**filter):
    return _sync_repo.find(**filter)


def find_one(**filter):
    return _sync_repo.find_one(**filter)


# -------------Pages CRUD-------------
def pages(**filter) -> Generator[Page, None, None]:
    return _sync_repo.pages(**filter)


def page(page_gid: str) -> Union[Page, None]:
    if page_gid is None:
        return None
    if not isinstance(page_gid, str):
        raise Exception
    return _sync_repo.page(page_gid)


def default_page() -> Page:
    return _sync_repo.default_page()


def insert_page(page_: Page) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.insert_page(page_)


def remove_page(page_: Page) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.remove_page(page_)


def update_page(page_: Page) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.update_page(page_)


# -------------Notes CRUD-------------


def insert_note(note_: Note, page: Page = None) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.insert_note(note_, page)


def update_note(note_: Note) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.update_note(note_)


def remove_note(note_: Note) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.remove_note(note_)


def notes(page_: Page | str):
    return _sync_repo.notes(page_)


def note(page_: Page | str, own_id: str):
    return _sync_repo.note(page_, own_id)


# -------------Arrow CRUD-------------
def insert_arrow(arrow_: Arrow) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.insert_arrow(arrow_)


def update_arrow(arrow_: Arrow) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.update_arrow(arrow_)


def remove_arrow(arrow_: Arrow) -> Change:
    fusion.call_delayed(_sync_repo.write_to_disk, 0)
    return _sync_repo.remove_arrow(arrow_)


def arrows(page_: Page | str):
    return _sync_repo.arrows(page_)


def arrow(page_: Page | str, own_id: str):
    return _sync_repo.arrow(page_, own_id)


# ------------For changes--------------
def apply_change(change: Change):
    return _sync_repo.apply_change(change)



from PySide6.QtCore import QRect, Qt
from PySide6.QtGui import QColor, QPaintEvent, QPainter, QResizeEvent
from PySide6.QtWidgets import QWidget

from fusion.libs.entity.change import Change
from fusion.platform.qt_widgets import bind_and_apply_state
from fusion.libs.state import view_state_type

import pamet
from pamet.desktop_app import default_note_font
from pamet.desktop_app.util import draw_text_lines, elide_text
from pamet.model.card_note import CardNote
from pamet.note_view_lib import register_note_view_type
from pamet.views.note.anchor.view_mixin import LinkNoteViewMixin
from pamet.views.note.base.view import NoteView
from pamet.views.note.base.state import NoteViewState
from pamet.views.note.image.image_label import ImageLabel
from pamet.views.note.qt_helpers import draw_link_decorations

from .ui_widget import Ui_CardNoteWidget


@view_state_type
class CardNoteViewState(NoteViewState, CardNote):
    pass


@register_note_view_type(state_type=CardNoteViewState,
                         note_type=CardNote,
                         edit=False)
class CardNoteWidget(QWidget, NoteView, LinkNoteViewMixin):

    def __init__(self, parent, initial_state):
        QWidget.__init__(self, parent)
        NoteView.__init__(self, parent=parent, initial_state=initial_state)
        LinkNoteViewMixin.__init__(self)

        self.ui = Ui_CardNoteWidget()
        self.ui.setupUi(self)

        self.setFont(default_note_font())
        self.image_label = ImageLabel(self)

        bind_and_apply_state(self, initial_state, self.on_state_change)
        self.destroyed.connect(lambda: self.disconnect_from_page_changes())

    def on_state_change(self, change: Change):
        state = change.last_state()

        if change.updated.color or change.updated.background_color:

            fg_col = QColor(*state.get_color().to_uint8_rgba_list())
            bg_col = QColor(*state.get_background_color().to_uint8_rgba_list())

            palette = self.palette()
            palette.setColor(self.backgroundRole(), bg_col)
            palette.setColor(self.foregroundRole(), fg_col)
            self.setPalette(palette)

        if change.updated.geometry:
            self.setGeometry(QRect(*state.geometry))

        if change.updated.local_image_url:
            self.image_label.update_image_cache(state.image_url,
                                                state.local_image_url)

        if change.updated.text or \
                change.updated.geometry \
                or change.updated.url:
            if '\n' in state.text:
                self._alignment = Qt.AlignLeft
            else:
                self._alignment = Qt.AlignHCenter

            self.recalculate_elided_text()

        if change.updated.url:
            url_page = pamet.page(state.url.get_page_id())
            if url_page and url_page.name:
                self.connect_to_page_changes(url_page.id)
            else:
                self.disconnect_from_page_changes()

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter()
        painter.begin(self)

        draw_text_lines(painter, self._elided_text_layout.data,
                        self._alignment,
                        self.state().text_rect())
        draw_link_decorations(self, painter)
        # Debug drawing
        # image_rect, text_rect = self.image_and_text_rects()
        # painter.setPen(Qt.red)
        # painter.drawRect(*image_rect.as_tuple())
        # painter.drawRect(*text_rect.as_tuple())
        painter.end()

    def resizeEvent(self, event: QResizeEvent):
        image_label_rect = QRect(*self.state().image_rect().as_tuple())
        self.image_label.setGeometry(image_label_rect)



from PySide6.QtCore import Qt
from PySide6.QtGui import QFocusEvent, QKeyEvent, QResizeEvent
from PySide6.QtWidgets import QListWidgetItem, QWidget
from fusion.view import View
import pamet

from thefuzz import process

from fusion.libs.command import commands
from fusion.libs.state import ViewState, view_state_type
from pamet.actions import tab as tab_actions
from pamet.actions import window as window_actions
from .ui_widget import Ui_CommandPaletteWidget


@view_state_type
class CommandPaletteViewState(ViewState):
    line_edit_text: str = ''


class CommandPaletteWidget(QWidget, View):

    def __init__(self, parent, initial_state: CommandPaletteViewState):
        QWidget.__init__(self, parent)
        View.__init__(self, initial_state)

        self.ui = Ui_CommandPaletteWidget()
        self.ui.setupUi(self)

        self.ui.lineEdit.setText(initial_state.line_edit_text)
        self.ui.lineEdit.textChanged.connect(self._handle_text_change)
        self.ui.lineEdit.setFocus()
        self.setMinimumHeight(self.ui.lineEdit.geometry().height())

        self.ui.suggestionsListWidget.itemActivated.connect(
            self._handle_item_activated)
        # # Close on Esc shortcut - moved to the Window.handle_esc_shortcut

        # Do the initial reuslts display
        self._handle_text_change(initial_state.line_edit_text)

    def focusOutEvent(self, event: QFocusEvent) -> None:
        window_actions.close_command_view(self.parent().state())

    def keyPressEvent(self, event: QKeyEvent) -> None:
        list_widget = self.ui.suggestionsListWidget
        if event.key() == Qt.Key_Up:
            list_widget.setCurrentRow(max(0, list_widget.currentRow() - 1))
        elif event.key() == Qt.Key_Down:
            list_widget.setCurrentRow(
                min(list_widget.count(),
                    list_widget.currentRow() + 1))
        elif event.key() == Qt.Key_Return:
            self._handle_item_activated(list_widget.currentItem())
        else:
            return super().keyPressEvent(event)

    def resizeEvent(self, event: QResizeEvent) -> None:
        self.adjustSize()
        return super().resizeEvent(event)

    def _handle_text_change(self, new_text: str):
        self.ui.suggestionsListWidget.clear()

        if new_text.startswith('>'):  # Show commands
            search_string = new_text[1:]
            choices_dict = {c: c.title for c in commands()}
            results = process.extract(search_string,
                                      choices=choices_dict,
                                      limit=None)

            for title, score, command in results:
                # if score
                item = QListWidgetItem(title)
                item.setData(Qt.UserRole, command)
                self.ui.suggestionsListWidget.addItem(item)

        else:  # No special prefix - show files to Go to
            choices_dict = {p.id: p.name for p in pamet.pages()}
            results = process.extract(new_text,
                                      choices=choices_dict,
                                      limit=None)

            tab = self.parent().current_tab()
            current_page = pamet.page(tab.state().page_view_state.page_id)

            for name, score, page_id in results:
                if page_id == current_page.id:
                    continue

                item = QListWidgetItem(name)
                page = pamet.page(page_id).copy()
                if not tab:
                    item.setData(Qt.UserRole,
                                 lambda p=page: window_actions.new_browser_tab(
                                     self.parent().state(), p))
                else:
                    item.setData(Qt.UserRole,
                                 lambda p=page: tab_actions.go_to_url(
                                     tab.state(), p.url()))

                self.ui.suggestionsListWidget.addItem(item)

        self.ui.suggestionsListWidget.setCurrentRow(0)
        self.adjustSize()

    def _handle_item_activated(self, item):
        if not item:
            window_actions.close_command_view(self.parent().state())
            return

        # Call the corresponding method (stored as item data)
        command = item.data(Qt.UserRole)
        window_actions.close_command_view(self.parent().state())
        command()



import pamet
from fusion.visual_inspection.action_camera import exec_action
from fusion import fsm

from pamet.model.page import Page
from fusion.util.point2d import Point2D


# Create new page
# Go back to the old page via the back link
# Go to the new page via the link
# Rename the page
# Go back to the old page via the Go to file command palette
# Go to the new page via the link
# Delete the new page
# Delete the old/initial page (to test the auto creation on last)
def test_test_page_crud(window_fixture, request):
    run_headless = request.config.getoption('--headless')

    with exec_action(delay_before_next=0.5894737243652344,
                     apply_delay=not run_headless):
        pamet.actions.tab.create_new_page(tab_state=fsm.view_state('2'),
                                          mouse_position=Point2D(940, 340))

    with exec_action(delay_before_next=0.017842769622802734,
                     apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('5'),
            new_size=Point2D(1838, 963))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('5'),
            new_size=Point2D(1570, 963))

    with exec_action(delay_before_next=0.0023813247680664062,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('5'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('5'),
            selection_updates_by_child={fsm.view_state('6'): True})

    with exec_action(delay_before_next=0.5292613506317139,
                     apply_delay=not run_headless):
        pamet.actions.tab.go_to_url(tab_state=fsm.view_state('2'),
                                    url="pamet:/p/00000001")

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('3'),
            new_size=Point2D(1838, 963))

    with exec_action(delay_before_next=0.0022859573364257812,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('8'): True})

    with exec_action(delay_before_next=0.526573657989502,
                     apply_delay=not run_headless):
        pamet.actions.tab.go_to_url(tab_state=fsm.view_state('2'),
                                    url="pamet:/p/00000005")

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('5'),
            new_size=Point2D(1838, 963))

    with exec_action(delay_before_next=0.5207712650299072,
                     apply_delay=not run_headless):
        pamet.actions.map_page.open_page_properties(
            tab_state=fsm.view_state('2'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('5'),
            new_size=Point2D(1570, 963))

    with exec_action(delay_before_next=0.0053997039794921875,
                     apply_delay=not run_headless):
        pamet.actions.map_page.save_page_properties(
            page=Page(id="00000005",
                      name="New page 1aaaa",
                      created="2022-09-09T09:54:42+03:00",
                      modified="2022-09-09T09:55:06+03:00"))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.tab.close_page_properties(tab_state=fsm.view_state('2'))

    with exec_action(delay_before_next=0.014553308486938477,
                     apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('5'),
            new_size=Point2D(1838, 963))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.apply_page_change_to_anchor_view(
            note_view_state=fsm.view_state('8'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.window.open_command_view(
            window_state=fsm.view_state('1'))

    with exec_action(delay_before_next=0.004543304443359375,
                     apply_delay=not run_headless):
        pamet.actions.window.close_command_view(
            window_state=fsm.view_state('1'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.tab.go_to_url(tab_state=fsm.view_state('2'),
                                    url="pamet:/p/00000001")

    with exec_action(delay_before_next=0.7721796035766602,
                     apply_delay=not run_headless):
        pamet.actions.window.open_command_view(
            window_state=fsm.view_state('1'))

    with exec_action(delay_before_next=0.0044956207275390625,
                     apply_delay=not run_headless):
        pamet.actions.window.close_command_view(
            window_state=fsm.view_state('1'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.tab.go_to_url(tab_state=fsm.view_state('2'),
                                    url="pamet:/p/00000005")

    with exec_action(delay_before_next=0.5170257091522217,
                     apply_delay=not run_headless):
        pamet.actions.map_page.open_page_properties(
            tab_state=fsm.view_state('2'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('5'),
            new_size=Point2D(1570, 963))

    with exec_action(delay_before_next=0.023119211196899414,
                     apply_delay=not run_headless):
        pamet.actions.map_page.delete_page(
            tab_view_state=fsm.view_state('2'),
            page=Page(id="00000005",
                      name="New page 1aaaa",
                      created="2022-09-09T09:54:42+03:00",
                      modified="2022-09-09T09:55:06+03:00"))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.tab.close_page_properties(tab_state=fsm.view_state('2'))

    with exec_action(delay_before_next=0.5176599025726318,
                     apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('3'),
            new_size=Point2D(1838, 963))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.apply_page_change_to_anchor_view(
            note_view_state=fsm.view_state('8'))

    with exec_action(delay_before_next=0.5212478637695312,
                     apply_delay=not run_headless):
        pamet.actions.map_page.open_page_properties(
            tab_state=fsm.view_state('2'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('3'),
            new_size=Point2D(1570, 963))

    with exec_action(delay_before_next=0.04296374320983887,
                     apply_delay=not run_headless):
        pamet.actions.map_page.delete_page(
            tab_view_state=fsm.view_state('2'),
            page=Page(id="00000001",
                      name="New page",
                      created="2022-09-09T09:54:30+03:00",
                      modified="2022-09-09T09:54:30+03:00"))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.tab.close_page_properties(tab_state=fsm.view_state('2'))

    with exec_action(delay_before_next=0.518906831741333,
                     apply_delay=not run_headless):
        pamet.actions.map_page.resize_page(
            map_page_view_state=fsm.view_state('14'),
            new_size=Point2D(1838, 963))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.apply_page_change_to_anchor_view(
            note_view_state=fsm.view_state('6'))

    assert True



from pathlib import Path
from typing import Union

from PySide6.QtGui import QImage

from fusion.util import get_new_id

from pamet.desktop_app.util import jpeg_blob_from_image
from pamet.util.url import Url
from slugify import slugify


class MediaStore:

    def __init__(self, media_root: Union[Path, str]):
        self.media_root = Path(media_root)

    def path_for_internal_uri(self, uri: Union[Url, str]) -> Union[Path, None]:
        uri = Url(uri)

        if not uri.is_internal():
            raise Exception('The given URI is not internal')

        cache_subfolder_name = uri.get_page_id()
        media_id = uri.get_media_id()
        return self.media_root / cache_subfolder_name / media_id

    def save_blob(self,
                  blob: bytes,
                  format: str = '',
                  cache_subfolder_name: str = '__common__',
                  source_uri: Union[Url, str] = None) -> Url:
        """Saves the given blob to a file in the media store.

        Each page has a separate folder. The file name consists of a slug
        extracted from the source_uri, the image id and an extension
        equal to the given format."""
        if not blob:
            return None

        # Generate the source_slug
        source_slug = ''
        if source_uri:
            source_uri = Url(source_uri)
            source_slug = Path(source_uri.path).name
            source_slug = slugify(source_slug, max_length=50)

        file_name = f'{source_slug}[{get_new_id()}]'
        if format:
            file_name += f'.{format}'
        file_path = self.media_root / cache_subfolder_name
        file_path.mkdir(parents=True, exist_ok=True)
        file_path = file_path / file_name

        try:
            bytes_written = file_path.write_bytes(blob)
        except Exception:
            return None

        if not bytes_written:
            return None

        return Url(f'pamet:///p/{cache_subfolder_name}/media/{file_name}')

    def save_image(self,
                   image: QImage,
                   format: str = 'jpg',
                   cache_subfolder_name: str = '__images__',
                   source_uri: Union[Url, str] = None) -> Url:
        if image.isNull():
            raise Exception('The given image is null')

        blob = jpeg_blob_from_image(image)
        return self.save_blob(blob, format, cache_subfolder_name, source_uri)



from __future__ import annotations

from dataclasses import field
from fusion.libs.entity import entity_type, Entity, get_entity_id


@entity_type
class PageChild(Entity):
    id: tuple[str, str] = field(default_factory=lambda: ('', get_entity_id()))

    @property
    def page_id(self):
        return self.id[0]

    @property
    def own_id(self):
        return self.id[1]

    def with_id(self, page_id: str = None, own_id: str = None):
        """A convinience method to produce a copy with a changed id (since
        the 'id' attribute is immutable (used in hashing))."""
        page_id = page_id or self.page_id
        own_id = own_id or self.own_id
        self_dict = self.asdict()
        self_dict['id'] = page_id, own_id
        return type(self)(**self_dict)

    @classmethod
    def in_page(cls,
                page: Page = None,
                page_id: str = None,
                own_id: str = None,
                **child_props):
        if page:
            if page_id:
                raise Exception('Only specify page OR page_id')
            page_id = page.id
        if page_id is None:
            raise Exception('You must specify either page or page_id.')

        own_id = own_id or get_entity_id()
        return cls(id=(page_id, own_id), **child_props)



from copy import copy
import time
from typing import Dict, List

from PySide6.QtWidgets import QWidget
from PySide6.QtCore import QPointF, Qt, QPoint, QTimer, QRectF
from PySide6.QtGui import QKeyEvent, QPainter, QPalette, QPicture, QImage, QColor, QBrush, QCursor
from PySide6.QtGui import QKeySequence, QShortcut

import fusion
from fusion.util import Point2D, Rectangle
from fusion.platform.qt_widgets import bind_and_apply_state
from fusion.view import View
from fusion import fsm as fsm

from pamet import commands
from pamet.constants import ARROW_EDGE_RAIDUS, MAX_RENDER_TIME
from pamet.constants import ALIGNMENT_LINE_LENGTH, RESIZE_CIRCLE_RADIUS
from pamet.constants import LONG_PRESS_TIMEOUT

import pamet
from pamet.desktop_app import selection_overlay_qcolor
from pamet.actions import map_page as map_page_actions
from pamet.desktop_app.mime_data_utils import entities_from_mime_data
from pamet.desktop_app.util import control_is_pressed
from pamet.model.arrow import Arrow, ArrowAnchorType
from pamet.model.note import Note
from pamet.util import snap_to_grid
from pamet.views.arrow.widget import ArrowView, ArrowViewState, ArrowWidget
from pamet.views.context_menu.widget import ContextMenuWidget
from pamet.views.map_page.view import MapPageView
from pamet.views.map_page.state import MapPageViewState, MapPageMode

from fusion.libs.entity.change import Change
from pamet.views.note.base.state import NoteViewState

log = fusion.get_logger(__name__)

IMAGE_CACHE_PADDING = 1
MIN_FULL_CHILD_RENDERS_PER_PAINT_EVENT = 10
DRAG_SELECT_COLOR = QColor(100, 100, 100, 50)


class NoteViewCache:

    def __init__(self):
        self.pcommand_cache: QPicture = None
        self.image_cache: QImage = None

        self.should_rebuild_pcommand_cache: bool = True
        self.should_reallocate_image_cache: bool = True
        self.should_rerender_image_cache: bool = True


def image_cache_rect_unprojected(display_rect: Rectangle):
    '''Project and round down to the pixel grid'''
    cache_rect = display_rect.toRect()

    cache_rect.moveTo(cache_rect.x() - IMAGE_CACHE_PADDING,
                      cache_rect.y() - IMAGE_CACHE_PADDING)
    cache_rect.setWidth(cache_rect.width() + 2 * IMAGE_CACHE_PADDING)
    cache_rect.setHeight(cache_rect.height() + 2 * IMAGE_CACHE_PADDING)

    return cache_rect


class MapPageWidget(QWidget, MapPageView):

    def __init__(self, parent, initial_state: MapPageViewState):
        QWidget.__init__(self, parent=parent)
        MapPageView.__init__(self, parent, initial_state)

        # Local variables
        self._cache_per_nc_view_id = {}
        self._paint_event_count = 0
        self.debug_drawing = False
        self._mouse_press_position = QPoint()
        self._note_widgets = {}
        self._arrow_widgets: Dict[str, ArrowWidget] = {}
        self._clipboard_outlines_shown = False

        # Setup shortcuts
        delete_shortcut = QShortcut(QKeySequence(Qt.Key_Delete), self)
        delete_shortcut.activated.connect(self.handle_delete_shortcut)

        color_notes_blue = QShortcut(QKeySequence('1'), self)
        color_notes_blue.activated.connect(
            lambda: map_page_actions.color_selected_children(
                self.state(),
                color=[0, 0, 1, 1],
                background_color=[0, 0, 1, 0.1]))
        color_notes_green = QShortcut(QKeySequence('2'), self)
        color_notes_green.activated.connect(
            lambda: map_page_actions.color_selected_children(
                self.state(),
                color=[0, 0.64, 0.235, 1],
                background_color=[0, 1, 0, 0.1]))
        color_notes_red = QShortcut(QKeySequence('3'), self)
        color_notes_red.activated.connect(
            lambda: map_page_actions.color_selected_children(
                self.state(),
                color=[1, 0, 0, 1],
                background_color=[1, 0, 0, 0.1]))
        color_notes_gray = QShortcut(QKeySequence('4'), self)
        color_notes_gray.activated.connect(
            lambda: map_page_actions.color_selected_children(
                self.state(),
                color=[0, 0, 0, 1],
                background_color=[0, 0, 0, 0.1]))
        remove_note_background = QShortcut(QKeySequence('5'), self)
        remove_note_background.activated.connect(
            lambda: map_page_actions.color_selected_children(
                self.state(), background_color=[0, 0, 0, 0]))
        select_all_shortcut = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_A), self)
        select_all_shortcut.activated.connect(
            lambda: map_page_actions.select_all_children(self.state()))
        QShortcut(QKeySequence(Qt.Key_E), self, commands.edit_selected_notes)
        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_E), self,
                  commands.open_page_properties)
        QShortcut(QKeySequence(Qt.Key_L), self, commands.start_arrow_creation)
        # QShortcut(QKeySequence(Qt.Key_Escape), self, self.handle_esc_shortcut,
        #           Qt.WidgetShortcut)
        QShortcut(
            QKeySequence(Qt.Key_A), self,
            lambda: map_page_actions.autosize_selected_notes(self.state()))
        QShortcut(QKeySequence('ctrl+Z'), self,
                  lambda: map_page_actions.undo(self.state()))
        QShortcut(QKeySequence('ctrl+shift+Z'), self,
                  lambda: map_page_actions.redo(self.state()))
        QShortcut(QKeySequence('ctrl+'), self,
                  lambda: map_page_actions.redo(self.state()))
        QShortcut(QKeySequence('ctrl+C'), self, commands.copy)
        QShortcut(QKeySequence('ctrl+V'), self, commands.paste)
        QShortcut(QKeySequence('ctrl+X'), self, commands.cut)
        QShortcut(QKeySequence('ctrl+shift+V'), self, commands.paste_special)
        QShortcut(QKeySequence('ctrl+Q'), self, commands.grab_screen_snippet)

        # Widget config
        pal = self.palette()
        pal.setColor(QPalette.ColorRole.Window, Qt.white)

        self.setPalette(pal)
        self.setMouseTracking(True)
        self.setFocusPolicy(Qt.ClickFocus)
        self.setAcceptDrops(True)
        self.setAutoFillBackground(True)

        self.subscriptions = []
        self.nw_subscribtions = {}  # by note widget
        self.arrow_view_state_subs_by_view_id = {}
        self.new_arrow_view_state_subs_by_state_id = {}

        # Subscribe to the page changes
        self.subscriptions.append(
            pamet.channels.entity_changes_by_id.subscribe(
                self.handle_page_change, index_val=self.state().page_id))
        # Subscribe to note changes
        self.subscriptions.append(
            pamet.channels.entity_changes_by_parent_gid.subscribe(
                self.handle_page_child_change, index_val=self.state().page_id))

        self.destroyed.connect(lambda: self.unsubscribe_all())

        bind_and_apply_state(self,
                             initial_state,
                             on_state_change=self.on_state_change)

    def dragEnterEvent(self, event):
        mime_data = event.mimeData()
        if mime_data.hasUrls() or mime_data.hasImage() or mime_data.hasText():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        mime_data = event.mimeData()
        entities = entities_from_mime_data(mime_data)

        q_mouse_pos = self.mapFromGlobal(QCursor.pos())
        mouse_pos = Point2D(q_mouse_pos.x(), q_mouse_pos.y())

        map_page_state = self.state()
        unproj_mouse_pos = map_page_state.unproject_point(mouse_pos)
        map_page_actions.insert_entities_relative_to(map_page_state, entities,
                                                     unproj_mouse_pos)

    def handle_page_change(self, change: Change):
        if change.updated.name:
            map_page_actions.handle_page_name_updated(self.parent_tab.state(),
                                                      change.last_state())

    def unsubscribe_all(self):
        for subscription in self.subscriptions:
            subscription.unsubscribe()

        for nw, subscription in self.nw_subscribtions.items():
            subscription.unsubscribe()

        for avs_id, sub in self.arrow_view_state_subs_by_view_id.items():
            sub.unsubscribe()

    def note_views(self) -> List[View]:
        yield from self.note_widgets()

    def arrow_views(self) -> List[ArrowView]:
        yield from self.arrow_widgets()

    def on_state_change(self, change: Change):
        state = change.last_state()
        if change.updated.viewport_height:
            # Invalidate image_cache for all children
            for child in self.note_widgets():
                nv_cache = self._note_widget_cache(child.view_id)
                nv_cache.should_reallocate_image_cache = True
                nv_cache.should_rerender_image_cache = True
            # self.update()

        if change.updated.special_mode:
            cursor: QCursor = self.cursor()
            if state.mode() == MapPageMode.CREATE_ARROW:
                cursor.setShape(Qt.CrossCursor)
            else:
                cursor.setShape(Qt.ArrowCursor)
            self.setCursor(cursor)

        for nv_state in change.added.note_view_states:
            self.add_note_widget(nv_state)

        for nv_state in change.removed.note_view_states:
            note_widget = self.note_widget(nv_state.view_id)
            if note_widget:
                # Don't fail if it's missing. When changing type - we delete
                # the note widget immediately (in order to avoid it handling
                # a note type it wasn't supposed to)
                self.remove_note_widget(note_widget)

        for av_state in change.added.arrow_view_states:
            self.add_arrow_widget(av_state)
        for av_state in change.removed.arrow_view_states:
            self.remove_arrow_widget(av_state)

        for av_state in change.added.new_arrow_view_states:
            self.add_arrow_widget(av_state)
        for av_state in change.removed.new_arrow_view_states:
            self.remove_arrow_widget(av_state)

        self.update()

    def add_arrow_widget(self, av_state: ArrowViewState):
        try:
            arrow_widget = ArrowWidget(av_state, self)
        except Exception as e:
            pamet.entity_broke(av_state, e)
            return

        self._arrow_widgets[av_state.view_id] = arrow_widget

        subscription = fsm.state_changes_per_TLA_by_view_id.subscribe(
            lambda change: self.handle_child_view_state_update(change),
            index_val=av_state.view_id)
        self.arrow_view_state_subs_by_view_id[av_state.view_id] = subscription

    def remove_arrow_widget(self, av_state: ArrowViewState):
        subscription = self.arrow_view_state_subs_by_view_id.pop(
            av_state.view_id)
        subscription.unsubscribe()
        arrow_widget = self._arrow_widgets.pop(av_state.view_id)
        arrow_widget.deleteLater()

    def add_note_widget(self, nv_state: NoteViewState):
        try:
            ViewType = pamet.note_view_type_by_state(type(nv_state).__name__)
            note_widget = ViewType(parent=self, initial_state=nv_state)
        except Exception as e:
            pamet.entity_broke(nv_state, e)
            return

        note_widget.hide()
        self._note_widgets[nv_state.view_id] = note_widget

        subscription = fsm.state_changes_per_TLA_by_view_id.subscribe(
            lambda change: self.handle_child_view_state_update(change),
            index_val=note_widget.view_id)
        self.nw_subscribtions[note_widget] = subscription
        self.update()

    def remove_note_widget(self, note_widget):
        nw_view_id = note_widget.view_id
        self.delete_note_view_cache(nw_view_id)
        subscription = self.nw_subscribtions.pop(note_widget)
        subscription.unsubscribe()
        note_widget = self._note_widgets.pop(nw_view_id)
        note_widget.deleteLater()
        self.update()

    def handle_child_view_state_update(self, change: Change):
        entity = change.last_state()
        if change.is_update():
            if isinstance(entity, NoteViewState):
                self.on_child_updated(change)
            else:  # If it's an arrow
                self.update()

        elif change.is_create():
            if isinstance(entity, NoteViewState):
                pass
                # self.on_child_updated(change)
            else:  # If it's an arrow
                self.update()

    def on_child_updated(self, change: Change):
        if not change.is_update():
            raise Exception

        nv_cache = self._note_widget_cache(change.new_state.view_id)

        if change.updated.geometry:
            if change.old_state.rect().size() != change.new_state.rect().size(
            ):
                nv_cache.should_rebuild_pcommand_cache = True
                nv_cache.should_reallocate_image_cache = True
                nv_cache.should_rerender_image_cache = True

        if change.updated.content:
            nv_cache.should_rebuild_pcommand_cache = True
            nv_cache.should_rerender_image_cache = True

        if change.updated.color or change.updated.background_color:
            nv_cache.should_rebuild_pcommand_cache = True
            nv_cache.should_rerender_image_cache = True

        self.update()

    def handle_page_child_change(self, change: Change):
        self_state = self.state()
        child = change.last_state()
        if change.is_create():
            map_page_actions.handle_child_added(self_state, child)

        elif change.is_delete():
            map_page_actions.handle_child_removed(self_state, child)

        else:  # Updated
            if isinstance(child, Note):
                if type(change.old_state) != type(change.new_state):
                    note_widget = self.note_widget_by_note_gid(child.gid())
                    self.remove_note_widget(note_widget)
                    map_page_actions.handle_child_removed(self_state, child)
                    map_page_actions.handle_child_added(self_state, child)
                    return

                child_widget = self.note_widget_by_note_gid(child.gid())
                child_widget_state = child_widget.state()
            elif isinstance(child, Arrow):
                child_widget_state = self.arrow_state_by_arrow_gid(child.gid())

            map_page_actions.handle_child_updated(child_widget_state, child)

    def arrow_state_by_arrow_gid(self, arrow_gid: str) -> ArrowViewState:
        for av_state in self.state().arrow_view_states:
            if av_state.gid() == arrow_gid:
                return av_state
        return None

    def _note_widget_cache(self, note_view_id):
        if note_view_id not in self._cache_per_nc_view_id:
            self._cache_per_nc_view_id[note_view_id] = NoteViewCache()

        return self._cache_per_nc_view_id[note_view_id]

    def delete_note_view_cache(self, note_view_id):
        if note_view_id not in self._cache_per_nc_view_id:
            return
        del self._cache_per_nc_view_id[note_view_id]

    def note_widgets(self):
        for nw_id, note_widget in self._note_widgets.items():
            yield note_widget

    def note_widget(self, view_Id):
        return self._note_widgets.get(view_Id, None)

    def note_widget_by_note_gid(self, note_gid):
        for nw_state_id, note_widget in self._note_widgets.items():
            if note_widget.state().note_gid == note_gid:
                return note_widget
        return None

    def arrow_widgets(self):
        for aw_id, arrow_widget in self._arrow_widgets.items():
            yield arrow_widget

    def arrow_widget(self, state_id: str):
        return self._arrow_widgets[state_id]

    def prep_command_cache_for_child(self, child):
        # In order to be pixel perfect - relay through QPicture
        # Otherwise QWidget rounds up coordinates.
        # RIP 30 hours of trying other workarounds
        pcommand_cache = QPicture()

        child.render(pcommand_cache, QPoint(0, 0))

        nv_cache = self._note_widget_cache(child.view_id)
        nv_cache.pcommand_cache = pcommand_cache
        nv_cache.should_rebuild_pcommand_cache = False

    def _render_image_cache_for_child(self, child: View,
                                      display_rect: Rectangle):
        painter = QPainter()

        cache_rect = image_cache_rect_unprojected(display_rect)
        nv_cache = self._note_widget_cache(child.view_id)

        pcommand_cache = nv_cache.pcommand_cache
        render_img = nv_cache.image_cache
        render_img.fill(0)

        if self.debug_drawing:
            render_img.fill(QColor(255, 0, 0, 100))

        painter.begin(render_img)

        # Drawing on the cache image needs to be
        # done with an offset = display_rect.x() % 1 otherwise there will be
        # visual artefats at the boundaries of adjacent notes.
        # RIP a few more hours of tracing those.

        draw_point_x = display_rect.x() - cache_rect.x()
        draw_point_y = display_rect.y() - cache_rect.y()

        scale_factor = self.state().height_scale_factor()

        painter.translate(draw_point_x, draw_point_y)
        painter.scale(scale_factor, scale_factor)

        pcommand_cache.play(painter)
        painter.end()

        nv_cache = self._note_widget_cache(child.view_id)
        nv_cache.image_cache = render_img
        nv_cache.should_rerender_image_cache = False

    def _draw_guide_lines_for_child(self, display_rect: Rectangle,
                                    painter: QPainter):
        # Draw guiding lines. How: get the four lines, sort the points,
        # elongate them and draw
        rect = display_rect
        h_lines = [[rect.top_left(), rect.top_right()],
                   [rect.bottom_left(),
                    rect.bottom_right()]]
        v_lines = [[rect.top_left(), rect.bottom_left()],
                   [rect.top_right(), rect.bottom_right()]]

        elong = ALIGNMENT_LINE_LENGTH
        all_lines = [*h_lines, *v_lines]

        for line in h_lines:
            p1, p2 = line

            p1 = Point2D(
                p1.x() - elong,  # p1.x == p2.x
                min(p1.y(), p1.y()))
            p2 = Point2D(p2.x() + elong, max(p1.y(), p2.y()))

            all_lines.append([p1, p2])

        for line in v_lines:
            p1, p2 = line

            p1 = Point2D(min(p1.x(), p2.x()), p1.y() - elong)  # p1.y == p2.y
            p2 = Point2D(max(p1.x(), p2.x()), p2.y() + elong)

            all_lines.append([p1, p2])

        for line in all_lines:
            p1, p2 = line
            points = [*p1.as_tuple(), *p2.as_tuple()]
            painter.drawLine(*points)

    def _visible_display_rects_by_view_id(self):
        viewport_rect = Rectangle(self.rect().x(),
                                  self.rect().y(),
                                  self.rect().width(),
                                  self.rect().height())
        self_state = self.state()
        unprojected_viewport = self_state.unproject_rect(viewport_rect)

        display_rects_by_child_view_id = {}  # {nc_id: rect}
        for note_widget in self.note_widgets():
            nt_rect = note_widget.state().rect()

            if not nt_rect.intersects(unprojected_viewport):
                continue

            unprojected_rect = self_state.project_rect(nt_rect)
            display_rects_by_child_view_id[note_widget.view_id] = QRectF(
                *unprojected_rect.as_tuple())

        return display_rects_by_child_view_id

    def paintEvent(self, event):
        state: MapPageViewState = self.state()
        self._paint_event_count += 1
        paint_t0 = time.time()

        painter = QPainter()
        painter.begin(self)

        painter.setRenderHint(QPainter.Antialiasing)
        painter.setRenderHint(QPainter.TextAntialiasing)

        pen = painter.pen()
        pen.setCosmetic(True)
        painter.setPen(pen)

        # Vars for stats
        placeholders_drawn = 0
        image_cache_memory_allocations = 0
        pc_cache_jobs = 0
        render_to_image_jobs = 0
        notes_reusing_cache = 0
        reused_expired_caches = 0
        non_urgent_count = 0
        paint_urgently = False

        # Determine the components inside the viewport
        display_rects_by_view_id = self._visible_display_rects_by_view_id()

        # Prep the unprojected mouse position to check if there's a note there
        q_mouse_pos = self.mapFromGlobal(self.cursor().pos())
        mouse_pos = Point2D(q_mouse_pos.x(), q_mouse_pos.y())
        unprojected_mouse_pos = self.state().unproject_point(mouse_pos)

        # Render the notes into a painter-commands-cache and an image cache
        # (prep only placeholders if slow)
        # At minimum renders one component into cache and preps placeholders
        for child_view_id, display_rect in display_rects_by_view_id.items():
            note_widget = self.note_widget(child_view_id)
            nw_cache = self._note_widget_cache(note_widget.view_id)

            if time.time() - paint_t0 > MAX_RENDER_TIME:
                paint_urgently = True

                if non_urgent_count >= MIN_FULL_CHILD_RENDERS_PER_PAINT_EVENT:
                    continue

            if nw_cache.should_rebuild_pcommand_cache:
                pc_cache_jobs += 1
                self.prep_command_cache_for_child(note_widget)

            # Allocate the image_cache. Done on resize, but not e.g. panning
            if nw_cache.should_reallocate_image_cache:
                image_cache_memory_allocations += 1
                cache_rect = image_cache_rect_unprojected(display_rect)

                render_img = QImage(cache_rect.size(),
                                    QImage.Format_ARGB32_Premultiplied)
                nw_cache.image_cache = render_img
                nw_cache.should_reallocate_image_cache = False

            if nw_cache.should_rerender_image_cache:
                render_to_image_jobs += 1
                non_urgent_count += 1

                self._render_image_cache_for_child(note_widget, display_rect)

            else:
                notes_reusing_cache += 1

        # Draw the cached images, along with various decorations
        for child_view_id, display_rect in display_rects_by_view_id.items():
            note_widget = self.note_widget(child_view_id)
            nw_cache = self._note_widget_cache(note_widget.view_id)
            note_vs = note_widget.state()

            render_img = nw_cache.image_cache

            nt_main_color = QColor(*note_vs.get_color().to_uint8_rgba_list())

            nt_background_color = QColor(
                *note_vs.get_background_color().to_uint8_rgba_list())

            nt_background_brush = QBrush(nt_background_color, Qt.SolidPattern)

            painter.setPen(nt_main_color)
            painter.setBrush(nt_background_brush)

            # Draw placeholders where the note has not been rendered
            if not render_img:
                placeholders_drawn += 1

                ph_color = QColor(nt_background_color)
                ph_color.setAlpha(60)
                ph_brush = QBrush(ph_color, Qt.DiagCrossPattern)

                painter.fillRect(display_rect, ph_brush)
                continue

            if nw_cache.should_rebuild_pcommand_cache or \
               nw_cache.should_reallocate_image_cache or \
               nw_cache.should_rerender_image_cache:
                reused_expired_caches += 1

            # Draw the cached redering of the component
            cache_rect = image_cache_rect_unprojected(display_rect)
            painter.drawImage(cache_rect, render_img)

            # Load test
            # for i in range(10):
            #     painter.drawImage(QRect(cache_rect.topLeft() + QPoint(i*50,i*50), cache_rect.size()) , render_img)

            # Draw directly from the command cache to the page
            # (several times slower)
            # painter.save()
            # painter.translate(cache_rect.topLeft())
            # painter.scale(
            #     self.height_scale_factor(),
            #     self.height_scale_factor())
            # nv_cache.pcommand_cache.play(painter)
            # # Load test
            # for i in range(10):
            #     painter.drawPicture(i*50,i*50, nv_cache.pcommand_cache)
            # painter.restore()

            # # Test text
            # painter.save()
            # font = painter.font()
            # font.setPointSizeF(20 * self.height_scale_factor())
            # painter.setFont(font)
            # painter.drawText(cache_rect.topLeft() + QPoint(0, 0.5), 'Test')
            # painter.restore()

            if self.debug_drawing:
                painter.save()
                painter.setPen(Qt.NoPen)
                green_brush = QBrush(Qt.green, Qt.SolidPattern)
                painter.fillRect(display_rect, green_brush)
                painter.restore()

            # Draw the selection overlay and resize circle for selected notes
            if note_widget.state() in state.selected_children:
                # Draw a yellow selection overlay
                painter.fillRect(display_rect, selection_overlay_qcolor)

                # Draw the resize circle
                center = display_rect.bottomRight()
                radius = RESIZE_CIRCLE_RADIUS
                radius *= self.state().height_scale_factor()

                circle_fill_color = QColor(nt_main_color)
                circle_fill_color.setAlpha(60)

                painter.save()
                painter.setBrush(QBrush(circle_fill_color, Qt.SolidPattern))

                painter.drawEllipse(center, radius, radius)
                painter.restore()

            # Draw lines for easier visual alignment
            if state.mode() in [
                    MapPageMode.NOTE_RESIZE, MapPageMode.CHILD_MOVE
            ]:
                if note_widget.state() in state.selected_children:
                    qdr = display_rect
                    dr = Rectangle(qdr.x(), qdr.y(), qdr.width(), qdr.height())
                    self._draw_guide_lines_for_child(dr, painter)

        # Draw arrows and their selection overlays
        painter.save()
        scale_factor = state.height_scale_factor()
        dx, dy = state.unproject_point(Point2D(0, 0)).as_tuple()

        painter.scale(scale_factor, scale_factor)
        painter.translate(-dx, -dy)

        # Draw the arrows
        for arrow_widget in self.arrow_widgets():
            draw_selection_overlay = False
            arrow_widget_state = arrow_widget.state()
            if (arrow_widget_state in state.selected_children
                    or arrow_widget_state in state.drag_selected_children):
                draw_selection_overlay = True

            draw_control_points = False
            if state.arrow_with_visible_cps:
                if state.arrow_with_visible_cps == arrow_widget_state:
                    draw_control_points = True

            arrow_widget.render(painter, draw_selection_overlay,
                                draw_control_points)

        painter.restore()
        # End of the arrow drawing

        # Draw the drag select rectangle
        if state.mode() == MapPageMode.DRAG_SELECT:
            # Draw the selection rects
            drag_select_rect = QRectF(*state.drag_select_rect_props)
            painter.fillRect(drag_select_rect, DRAG_SELECT_COLOR)

            # Draw the selection overlays
            for note_view_state in state.drag_selected_children:
                if note_view_state.view_id not in display_rects_by_view_id:
                    continue
                painter.fillRect(
                    display_rects_by_view_id[note_view_state.view_id],
                    selection_overlay_qcolor)
                # elif nc_id in self._arrow_widgets:
                # ^^ that gets done while drawing the arrows

        elif state.mode() in [
                MapPageMode.CREATE_ARROW, MapPageMode.ARROW_EDGE_DRAG
        ]:
            # Draw the anchors for the notes under/close to the mouse
            for note_widget in self.get_note_views_at(mouse_pos,
                                                      ARROW_EDGE_RAIDUS):
                note_vs = note_widget.state()
                for anchor_type in ArrowAnchorType.real_types():
                    arr_anchor_pos = note_vs.arrow_anchor(anchor_type)
                    arr_anchor_pos = state.project_point(arr_anchor_pos)
                    radius = ARROW_EDGE_RAIDUS * state.height_scale_factor()

                    q_anchor_pos = QPointF(*arr_anchor_pos.as_tuple())
                    nt_main_color = QColor(
                        *note_vs.get_color().to_uint8_rgba_list())

                    nt_background_color = QColor(
                        *note_vs.get_background_color().to_uint8_rgba_list())

                    nt_background_brush = QBrush(nt_background_color,
                                                 Qt.SolidPattern)

                    painter.setPen(nt_main_color)
                    painter.setBrush(nt_background_brush)

                    # q_proj_mouse = QPointF(
                    #     *state.project_point(unprojected_mouse_pos).as_tuple())
                    # painter.drawEllipse(q_proj_mouse, 100, 100)
                    # painter.drawEllipse(q_mouse_pos, 100, 100)
                    painter.drawEllipse(q_anchor_pos, radius, radius)

        # Draw the clipboard outlines if control is pressed
        if control_is_pressed():
            self._clipboard_outlines_shown = True
            for note in pamet.clipboard.get_contents():
                if not isinstance(note, Note):
                    continue
                nt_main_color = QColor(*note.get_color().to_uint8_rgba_list())
                painter.setPen(nt_main_color)
                painter.setBrush(Qt.NoBrush)
                rect = note.rect()
                rect.set_top_left(
                    snap_to_grid(unprojected_mouse_pos + rect.top_left()))
                rect = state.project_rect(rect)
                painter.drawRect(QRectF(*rect.as_tuple()))

        # Report stats
        notes_on_screen = len(display_rects_by_view_id)

        latency = 1000 * (time.time() - paint_t0)
        # fps = 1000 / latency

        painter.setPen(Qt.red)
        stats_text = (
            'Last render stats: %s notes on screen ,'
            ' %s newly allocated,'
            ' %s as placeholders, %s PC cache jobs, '
            '%s render to img jobs, %s reusing cache, '
            '%s reusing expired.Latency: %sms.'
            ' Painted urgently: %s. Paint event count: %s' %
            (notes_on_screen, image_cache_memory_allocations,
             placeholders_drawn, pc_cache_jobs, render_to_image_jobs,
             notes_reusing_cache, reused_expired_caches, int(latency),
             'True' if paint_urgently else 'False', self._paint_event_count))

        log.info(stats_text)
        painter.drawText(10, 10, stats_text)
        painter.end()

        if paint_urgently and \
           (reused_expired_caches > 0 or placeholders_drawn > 0):
            QTimer.singleShot(0, self.update)

    def check_for_long_press(self):
        new_pos = self.mapFromGlobal(QCursor.pos())
        if self._mouse_press_position == new_pos and \
           self._left_mouse_is_pressed:
            self.handle_left_mouse_long_press(Point2D(new_pos.x(),
                                                      new_pos.y()))

    def mousePressEvent(self, event):
        self._mouse_press_position = event.pos()
        mouse_pos = Point2D(event.pos().x(), event.pos().y())
        state = self.state()

        QTimer.singleShot(LONG_PRESS_TIMEOUT * 1000, self.check_for_long_press)

        if event.button() is Qt.LeftButton:

            # If there's an arrow selected
            # check if an arrow edge is under the mouse

            if state.arrow_with_visible_cps:
                arrow_widget = self.arrow_widget(
                    state.arrow_with_visible_cps.view_id)

                edge_under_mouse_idx = arrow_widget.edge_at(mouse_pos)
                if edge_under_mouse_idx is not None:
                    map_page_actions.start_arrow_edge_drag(
                        state, mouse_pos, edge_under_mouse_idx)
                    return

            self.handle_left_mouse_press(mouse_pos)

        elif event.button() is Qt.RightButton:
            self.handle_right_mouse_press(
                Point2D(event.pos().x(),
                        event.pos().y()))

    def mouseReleaseEvent(self, event):
        mouse_pos = Point2D(event.pos().x(), event.pos().y())
        if event.button() is Qt.LeftButton:
            self.handle_left_mouse_release(mouse_pos)

        elif event.button() is Qt.MiddleButton:
            self.middle_click_event(mouse_pos)

    def mouseMoveEvent(self, event):
        old_nvs_under_mouse = copy(self._note_views_under_mouse)
        self.handle_mouse_move(Point2D(event.pos().x(), event.pos().y()))

        if self._note_views_under_mouse != old_nvs_under_mouse:
            if self.state().mode() == MapPageMode.CREATE_ARROW:
                self.update()

        # Prompt updates for the clipboard outline visualization
        if control_is_pressed():
            self.update()

    def mouseDoubleClickEvent(self, event):
        mouse_pos = Point2D(event.pos().x(), event.pos().y())
        state = self.state()

        if event.button() is Qt.LeftButton:
            if state.arrow_with_visible_cps:
                arrow_widget = self.arrow_widget(
                    state.arrow_with_visible_cps.view_id)

                edge_under_mouse_idx = arrow_widget.edge_at(mouse_pos)
                if edge_under_mouse_idx is not None:
                    map_page_actions.delete_arrow_edge(
                        state.arrow_with_visible_cps, edge_under_mouse_idx)
                    return

            self.left_mouse_double_click_event(
                Point2D(event.pos().x(),
                        event.pos().y()))

    def wheelEvent(self, event):
        degrees = event.angleDelta().y() / 8
        steps = degrees / 15

        self.handle_mouse_scroll(steps)

    def resizeEvent(self, event):
        self.update()
        new_size = Point2D(event.size().width(), event.size().height())
        self.handle_resize_event(new_size)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        # For the clipboard outline visualizations
        if event.key() == Qt.Key_Control:
            self.update()
        return super().keyPressEvent(event)

    def keyReleaseEvent(self, event: QKeyEvent) -> None:
        # For the clipboard outline visualizations
        if event.key() == Qt.Key_Control:
            self.update()
        return super().keyReleaseEvent(event)

    def handle_right_mouse_press(self, position):
        ncs_under_mouse = list(self.get_note_views_at(position))
        menu_entries = {}

        mode = self.state().mode()

        if mode == MapPageMode.NONE:
            if ncs_under_mouse:
                map_page_actions.update_child_selections(
                    self.state(), {nv.state(): True
                                   for nv in ncs_under_mouse})
                menu_entries['Edit note'] = commands.edit_selected_notes
                menu_entries['Copy'] = commands.copy
                if pamet.clipboard.get_contents():
                    menu_entries['Paste'] = commands.paste
                menu_entries['Cut'] = commands.cut
            else:
                menu_entries['New note'] = commands.create_new_note

            menu_entries['Paste special'] = commands.paste_special
            menu_entries['New page'] = commands.create_new_page
            menu_entries['Create arrow'] = commands.start_arrow_creation

        else:
            if mode == MapPageMode.CREATE_ARROW:
                text = 'Cancel arrow creation'
                menu_entries[
                    text] = lambda: map_page_actions.abort_special_mode(
                        self.state())
            else:
                # If we're in another special mode - don't show the menu
                return

        # Open the context menu with the tab as its parent
        context_menu = ContextMenuWidget(self.parent(), entries=menu_entries)
        context_menu.popup_on_mouse_pos()

    def handle_esc_shortcut(self):
        if self.state().mode() != MapPageMode.NONE:
            map_page_actions.abort_special_mode(self.state())

    def update_undo_redo_buttons_enabled(self):
        pass



from pathlib import Path
from fusion.libs.entity import entity_type
from pamet.model.text_note import TextNote

# SCRIPT = 'script'
SCRIPT_PATH = 'script_path'
COMMAND_ARGS = 'command_args'


@entity_type
class ScriptNote(TextNote):
    run_in_terminal: bool = True

    @property
    def script_path(self) -> Path | None:
        path_str = self.content.get(SCRIPT_PATH, '')
        return Path(path_str) if path_str else None

    @script_path.setter
    def script_path(self, script: Path | str):
        if script is None:
            self.content.pop(SCRIPT_PATH, None)
            return
        self.content[SCRIPT_PATH] = str(script)

    @property
    def command_args(self) -> str:
        return self.content.get(COMMAND_ARGS, '')

    @command_args.setter
    def command_args(self, script: str):
        self.content[COMMAND_ARGS] = script



from __future__ import annotations

from fusion import fsm
from fusion.libs.action import action
from fusion.libs.state import ViewState
from pamet.model.page import Page
from pamet.views.command_palette.widget import CommandPaletteViewState
from pamet.views.tab.state import TabViewState

from pamet.actions import tab as tab_actions
from pamet.views.window.state import WindowViewState


@action('window.new_browser_window')
def new_browser_window():
    window_state = WindowViewState()
    fsm.add_state(window_state)
    return window_state


@action('window.new_browser_tab')
def new_browser_tab(window_state: WindowViewState, page: Page):
    tab_state = TabViewState()
    fsm.add_state(tab_state)

    window_state.tab_states.append(tab_state)
    fsm.update_state(window_state)

    tab_actions.go_to_url(tab_state, page.url())


@action('window.close_tab')
def close_tab(window_state: WindowViewState, tab_state: TabViewState):
    window_state.tab_states.remove(tab_state)
    fsm.update_state(window_state)


@action('window.open_command_view')
def open_command_view(window_state: WindowViewState,
                      prefix: str = '',
                      command_view_state: ViewState = None):

    if not command_view_state:
        command_view_state = CommandPaletteViewState(line_edit_text=prefix)
        fsm.add_state(command_view_state)

    window_state.command_view_state = command_view_state
    fsm.update_state(window_state)


@action('window.close_command_view')
def close_command_view(window_state: WindowViewState):
    window_state.command_view_state = None
    fsm.update_state(window_state)



from dataclasses import field
from pathlib import Path

from PySide6.QtCore import QStandardPaths

from fusion import get_logger
from fusion.libs.entity import entity_type
from pamet.config import PametSettings

log = get_logger(__name__)

user_data_path = QStandardPaths.writableLocation(
    QStandardPaths.GenericDataLocation)
pamet_data_folder_path = Path(user_data_path) / 'pamet'
pamet_data_folder_path.mkdir(parents=True, exist_ok=True)


@entity_type
class UserDesktopSettings(PametSettings):

    repository_path: str = pamet_data_folder_path / 'repo'
    media_store_path: str = pamet_data_folder_path / 'media_store'
    scripts_folder: str = pamet_data_folder_path / 'scripts'
    script_templates_folder: str = pamet_data_folder_path / 'script_templates'
    accepted_script_risks: bool = False
    run_in_terminal_prefix_posix: str = 'gnome-terminal -- '
    run_in_terminal_prefix_windows: str = 'powershell -noexit '


@entity_type
class RepoSettings(PametSettings):
    repo_path: Path = field(repr=False,
                            default=None)
    backups_enabled: bool = True
    backup_folder: str = None
    home_page: str = None
    record_all_changes: bool = False
    semantic_search_enabled: bool = False

    def __post_init__(self):
        if self.repo_path is None:
            raise Exception

        if not self.backup_folder:
            self.backup_folder = self.repo_path / '.pamet' / 'backups'



from thefuzz import fuzz

from pamet.services.search.base import BaseSearchService, SearchResult


class FuzzySearchService(BaseSearchService):

    def text_search(self, text: str) -> SearchResult:
        """Searches the cached texts and returns search results"""
        search_results = []
        text = text.lower()
        for note_gid, index_entry in self.index.items():
            score = 0
            if index_entry.content_lowered.startswith(text):
                score = 1
            elif index_entry.content_lowered.find(text) != -1:
                score = 0.99
            else:
                score = fuzz.ratio(text, index_entry.content_lowered)
                score = (score / 100) * 0.98

            if score <= 0.4:
                continue

            search_result = SearchResult(
                note_gid,
                content_string=index_entry.content_string,
                score=score)
            search_results.append(search_result)

        search_results = sorted(search_results,
                                key=lambda r: r.score,
                                reverse=True)
        return search_results




from copy import copy
from typing import Tuple
from fusion.util.point2d import Point2D
from fusion.util.rectangle import Rectangle
from fusion.libs.entity import entity_type
from pamet.constants import NOTE_MARGIN
from pamet.model.image_note import ImageNote
from pamet.model.text_note import TextNote

MIN_AR_DELTA_FOR_HORIZONTAL_ALIGN = 0.5
IMAGE_PORTION_FOR_HORIZONTAL_ALIGN = 0.8


@entity_type
class CardNote(ImageNote, TextNote):
    def image_and_text_rects(self,
                             for_size: Point2D = None
                             ) -> Tuple[Rectangle, Rectangle]:
        """Returns the rectangles where the image and text should fit.
        They are relative to the note view"""
        image_rect = Rectangle(0, 0, 0, 0)
        text_rect = Rectangle(0, 0, 0, 0)

        # Fit the image in the note
        # If there's no image - assume rectangular
        image_size = self.image_size
        if image_size:
            image_aspect_ratio = image_size.x() / image_size.y()
        else:
            image_aspect_ratio = 1

        if for_size:
            note_size = for_size
        else:
            note_size = self.size()
        note_aspect_ratio = note_size.x() / note_size.y()
        ar_delta = note_aspect_ratio - image_aspect_ratio
        if ar_delta > MIN_AR_DELTA_FOR_HORIZONTAL_ALIGN:
            # Calc image size from AR
            image_rect.set_size(
                Point2D(note_size.y() * image_aspect_ratio, note_size.y()))
            # Prep the field (will apply margins below)
            text_field = Rectangle(*image_rect.top_right().as_tuple(),
                                   note_size.x() - image_rect.width(),
                                   note_size.y())

        else:
            # Decide on a good image to text ratio. E.g. 5:1 vertically
            # Image rect
            image_field = Rectangle(
                0, 0, note_size.x(),
                note_size.y() * IMAGE_PORTION_FOR_HORIZONTAL_ALIGN)
            image_field_AR = image_field.width() / image_field.height()
            if image_field_AR > image_aspect_ratio:
                # The avalable space is wider than the image
                image_rect.set_size(
                    Point2D(image_field.height() * image_aspect_ratio,
                            image_field.height()))
            else:
                # The available space is higher (when we fit the image width)
                image_rect.set_size(
                    Point2D(image_field.width(),
                            image_field.width() / image_aspect_ratio))
                image_field = copy(image_rect)

            image_rect.move_center(image_field.center())

            # Text rect
            text_field = Rectangle(image_field.bottom_left().x(),
                                   image_field.bottom_left().y(),
                                   note_size.x(),
                                   note_size.y() - image_field.height())

        # Apply margins to get the text field
        text_rect = text_field
        text_rect.set_top_left(text_rect.top_left() +
                               Point2D(NOTE_MARGIN, NOTE_MARGIN))
        text_rect.set_size(text_rect.size() -
                           Point2D(2 * NOTE_MARGIN, 2 * NOTE_MARGIN))

        return image_rect, text_rect

    def image_rect(self, for_size: Point2D = None):
        return self.image_and_text_rects(for_size)[0]

    def text_rect(self, for_size: Point2D = None):
        return self.image_and_text_rects(for_size)[1]



animals = '''Ant
Antelope
Baboon
Bat
Beagle
Bear
Bird
Butterfly
Cat
Caterpillar
Chicken
Cow
Dog
Dolphin
Donkey
Eagle
Fish
Fly
Fox
Frog
Gerbil
Goose
Gopher
Gorilla
Heron
Honey Bee
Horn Shark
Horse
Ibis
Iguana
Impala
Jackal
Jaguar
Javanese
Jellyfish
Kakapo
Kangaroo
King Penguin
Kiwi
Koala
Lemming
Lemur
Leopard
Saola
Scorpion
Snake
Swan
Tuatara
Turkey
Zebra'''.split('\n')

fruits = '''Apple
Banana
Orange
Mango
Grapes
Pineapple
Pomegranate
Avocado
Coconut
Papaya
Watermelon
Dragonfruit
Strawberry
Blueberry
Blackberry
Gooseberry
Cherry
Apricot
Jackfruit
Kiwifruit
Lime
Peach
Pear
Tamarind
Custard-Apple
Fig
Guava
MuskMelon
Olives
Java Plum'''.split('\n')

adjectives = '''able
best
better
big
black
certain
clear
different
early
easy
economic
federal
free
full
good
great
grumpy
gifted
hard
high
human
hilarious
important
international
large
late
little
local
long
low
major
new
only
other
possible
public
real
recent
right
small
social
special
strong
sure
true
white
whole'''.split('\n')



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'widget.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QHBoxLayout, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_BaseNoteEditWidget(object):
    def setupUi(self, BaseNoteEditWidget):
        if not BaseNoteEditWidget.objectName():
            BaseNoteEditWidget.setObjectName(u"BaseNoteEditWidget")
        BaseNoteEditWidget.resize(640, 480)
        self.verticalLayout = QVBoxLayout(BaseNoteEditWidget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.topLayout = QHBoxLayout()
        self.topLayout.setObjectName(u"topLayout")
        self.topLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.topLayout.addItem(self.horizontalSpacer)

        self.toolbarLayout = QHBoxLayout()
        self.toolbarLayout.setObjectName(u"toolbarLayout")
        self.toolbarLayout.setContentsMargins(6, 6, 6, 6)
        self.devButton = QPushButton(BaseNoteEditWidget)
        self.devButton.setObjectName(u"devButton")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.devButton.sizePolicy().hasHeightForWidth())
        self.devButton.setSizePolicy(sizePolicy)
        self.devButton.setMinimumSize(QSize(0, 0))

        self.toolbarLayout.addWidget(self.devButton)


        self.topLayout.addLayout(self.toolbarLayout)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.topLayout.addItem(self.horizontalSpacer_2)


        self.verticalLayout.addLayout(self.topLayout)

        self.centralAreaWidget = QWidget(BaseNoteEditWidget)
        self.centralAreaWidget.setObjectName(u"centralAreaWidget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.centralAreaWidget.sizePolicy().hasHeightForWidth())
        self.centralAreaWidget.setSizePolicy(sizePolicy1)
        self.verticalLayout_3 = QVBoxLayout(self.centralAreaWidget)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)

        self.verticalLayout.addWidget(self.centralAreaWidget)

        self.bottomLayout = QHBoxLayout()
        self.bottomLayout.setObjectName(u"bottomLayout")
        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.bottomLayout.addItem(self.horizontalSpacer_3)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setSpacing(9)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(6, 6, 6, 6)
        self.cancelButton = QPushButton(BaseNoteEditWidget)
        self.cancelButton.setObjectName(u"cancelButton")

        self.horizontalLayout_4.addWidget(self.cancelButton)

        self.saveButton = QPushButton(BaseNoteEditWidget)
        self.saveButton.setObjectName(u"saveButton")

        self.horizontalLayout_4.addWidget(self.saveButton)


        self.bottomLayout.addLayout(self.horizontalLayout_4)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.bottomLayout.addItem(self.horizontalSpacer_4)


        self.verticalLayout.addLayout(self.bottomLayout)


        self.retranslateUi(BaseNoteEditWidget)

        QMetaObject.connectSlotsByName(BaseNoteEditWidget)
    # setupUi

    def retranslateUi(self, BaseNoteEditWidget):
        BaseNoteEditWidget.setWindowTitle(QCoreApplication.translate("BaseNoteEditWidget", u"BaseNoteEditWidget", None))
        self.devButton.setText(QCoreApplication.translate("BaseNoteEditWidget", u"dev", None))
        self.cancelButton.setText(QCoreApplication.translate("BaseNoteEditWidget", u"Cancel (Esc)", None))
        self.saveButton.setText(QCoreApplication.translate("BaseNoteEditWidget", u"Save (ctrl+s)", None))
#if QT_CONFIG(shortcut)
        self.saveButton.setShortcut(QCoreApplication.translate("BaseNoteEditWidget", u"Ctrl+S", None))
#endif // QT_CONFIG(shortcut)
    # retranslateUi




from __future__ import annotations
from copy import copy
import json
from pathlib import Path
from PySide6.QtGui import QColor

from fusion.logging import get_logger

from pamet.constants import SELECTION_OVERLAY_COLOR
from pamet.desktop_app.app import DesktopApp
from pamet.desktop_app.config import RepoSettings, UserDesktopSettings
from pamet.desktop_app.icon_cache import PametQtWidgetsCachedIcons
from pamet.services.script_runner import ScriptRunner
from pamet.desktop_app.config import pamet_data_folder_path

log = get_logger(__name__)

SETTINGS_JSON = 'settings.json'

icons = PametQtWidgetsCachedIcons()

selection_overlay_qcolor = QColor(
    *SELECTION_OVERLAY_COLOR.to_uint8_rgba_list())

_media_store = None
_default_note_font = None
_app = None
script_runner = ScriptRunner()


def user_settings_path() -> Path:
    return pamet_data_folder_path / SETTINGS_JSON


def repo_settings_path(repo_path: Path) -> Path:
    return repo_path / '.pamet' / SETTINGS_JSON


# Config handling
def get_user_settings() -> UserDesktopSettings:
    settings_path = user_settings_path()
    if not settings_path.exists():
        settings = UserDesktopSettings()
        save_user_settings(settings)
        return settings

    with open(settings_path) as settings_file:
        config_dict = json.load(settings_file)
        return UserDesktopSettings.load(config_dict)


def save_user_settings(updated_config: UserDesktopSettings):
    config_str = json.dumps(updated_config.asdict(),
                            indent=4,
                            ensure_ascii=False)
    config_path = user_settings_path()
    config_path.write_text(config_str)


def get_repo_settings(repo_path: Path) -> RepoSettings:
    config_path = repo_settings_path(repo_path)
    if not config_path.exists():
        settings = RepoSettings(repo_path=repo_path)
        save_repo_settings(settings)
        return settings

    with open(config_path) as config_file:
        config_dict = json.load(config_file)
        dict_on_load = copy(config_dict)
        settings_id = config_dict.pop('id')
        settings = RepoSettings(id=settings_id, repo_path=repo_path)
        settings.replace_silent(**config_dict)
        settings._dict_on_load = dict_on_load
        return settings


def save_repo_settings(repo_settings: RepoSettings):
    config_str = json.dumps(repo_settings.asdict(),
                            indent=4,
                            ensure_ascii=False)
    config_path = repo_settings_path(repo_settings.repo_path)
    config_path.parent.mkdir(parents=True, exist_ok=True)
    config_path.write_text(config_str)


def media_store():
    return _media_store


def set_media_store(new_media_store):
    global _media_store
    _media_store = new_media_store


def default_note_font():
    return copy(_default_note_font)


def set_default_note_font(new_default_note_font):
    global _default_note_font
    _default_note_font = new_default_note_font


def get_app() -> DesktopApp:
    return _app


def set_app(new_app):
    global _app
    _app = new_app



from dataclasses import field
from enum import Enum
from typing import List
from fusion.util.color import Color
from fusion.util.point2d import Point2D

from fusion.libs.entity import entity_type
from fusion.logging import get_logger

import pamet
from pamet.constants import DEFAULT_ARROW_THICKNESS
from pamet.constants import DEFAULT_BG_COLOR, DEFAULT_COLOR
from pamet.model.page_child import PageChild

log = get_logger(__name__)
BEZIER_CUBIC = 'bezier_cubic'


class ArrowAnchorType(Enum):
    NONE = 0
    AUTO = 1
    MID_LEFT = 2
    TOP_MID = 3
    MID_RIGHT = 4
    BOTTOM_MID = 5

    @classmethod
    def real_types(cls):
        yield from [cls.MID_LEFT, cls.TOP_MID, cls.MID_RIGHT, cls.BOTTOM_MID]


@entity_type
class Arrow(PageChild):
    tail_coords: list = field(default_factory=list)
    tail_note_id: str = None
    tail_anchor: str = ArrowAnchorType.NONE.name

    mid_point_coords: List[list] = field(default_factory=list)

    head_coords: list = field(default_factory=list)
    head_note_id: str = None
    head_anchor: str = ArrowAnchorType.NONE.name

    color: tuple = DEFAULT_COLOR
    # background_color: tuple = DEFAULT_BG_COLOR
    line_type: str = None
    line_thickness: float = DEFAULT_ARROW_THICKNESS
    line_function_name: str = BEZIER_CUBIC
    head_shape: str = None
    tail_shape: str = None

    def parent_gid(self):
        return self.page_id

    def get_parent_page(self):
        return pamet.page(self.page_id)

    @property
    def tail_point(self) -> Point2D:
        if not self.tail_coords:
            return None
        return Point2D(*self.tail_coords)

    @tail_point.setter
    def tail_point(self, point: Point2D):
        if point:
            self.tail_coords = point.as_tuple()
        else:
            self.tail_coords = None

    @property
    def head_point(self) -> Point2D:
        if not self.head_coords:
            return None
        return Point2D(*self.head_coords)

    @head_point.setter
    def head_point(self, point: Point2D):
        if point:
            self.head_coords = point.as_tuple()
        else:
            self.head_coords = None

    @property
    def mid_points(self) -> List[Point2D]:
        return [Point2D(*mid_point) for mid_point in self.mid_point_coords]

    def get_midpoint(self, idx: int) -> Point2D:
        return Point2D(*self.mid_point_coords[idx])

    def replace_midpoints(self, midpoint_list: List[Point2D]):
        self.mid_point_coords = [mp.as_tuple() for mp in midpoint_list]

    def get_color(self) -> Color:
        return Color(*self.color)

    def set_color(self, color: Color):
        self.color = color.as_tuple()

    @property
    def tail_anchor_type(self):
        return ArrowAnchorType[self.tail_anchor]

    @tail_anchor_type.setter
    def tail_anchor_type(self, new_type: ArrowAnchorType):
        self.tail_anchor = new_type.name

    @property
    def head_anchor_type(self):
        return ArrowAnchorType[self.head_anchor]

    @head_anchor_type.setter
    def head_anchor_type(self, new_type: ArrowAnchorType):
        self.head_anchor = new_type.name

    def has_tail_anchor(self):
        return bool(self.tail_note_id)

    def has_head_anchor(self):
        return bool(self.head_note_id)

    def edge_indices(self):
        mid_edge_count = 2 + len(self.mid_points)
        return list(range(mid_edge_count))

    def potential_edge_indices(self):
        return [i + 0.5 for i in self.edge_indices()[:-1]]

    def all_edge_indices(self):
        return sorted(self.edge_indices() + self.potential_edge_indices())

    def set_tail(self,
                 fixed_pos: Point2D = None,
                 anchor_note_id: str = None,
                 anchor_type: ArrowAnchorType = ArrowAnchorType.NONE):
        if fixed_pos and anchor_note_id:
            # The fixed pos is almost always propagated
            # but if there's an anchor note - it takes precedence
            fixed_pos = None

        if fixed_pos and anchor_type != ArrowAnchorType.NONE:
            raise Exception

        self.tail_point = fixed_pos
        self.tail_note_id = anchor_note_id
        self.tail_anchor_type = anchor_type

    def set_head(self,
                 fixed_pos: Point2D = None,
                 anchor_note_id: str = None,
                 anchor_type: ArrowAnchorType = ArrowAnchorType.NONE):
        if fixed_pos and anchor_note_id:
            # The fixed pos is almost always propagated
            # but if there's an anchor note - it takes precedence
            fixed_pos = None

        if fixed_pos and anchor_type != ArrowAnchorType.NONE:
            raise Exception

        self.head_point = fixed_pos
        self.head_note_id = anchor_note_id
        self.head_anchor_type = anchor_type
